import csv
import json
import logging
import os

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.signal import unit_impulse
from tqdm import tqdm

from clarity.evaluator.mbstoi.mbstoi import mbstoi
from clarity.evaluator.mbstoi.mbstoi_utils import find_delay_impulse
from clarity.evaluator.msbg.audiogram import Audiogram
from clarity.evaluator.msbg.msbg import Ear
from clarity.evaluator.msbg.msbg_utils import MSBG_FS, pad, read_signal, write_signal


def listen(ear, signal, audiogram_l, audiogram_r):
    """
    Generate MSBG processed signal
    :param ear: MSBG ear
    :param wav: binaural signal
    :return: binaural signal
    """
    ear.set_audiogram(audiogram_l)
    out_l = ear.process(signal[:, 0])

    ear.set_audiogram(audiogram_r)
    out_r = ear.process(signal[:, 1])
    if len(out_l[0]) != len(out_r[0]):
        diff = len(out_l[0]) - len(out_r[0])
        if diff > 0:
            out_r[0] = np.flipud(pad(np.flipud(out_r[0]), len(out_l[0])))
        else:
            out_l[0] = np.flipud(pad(np.flipud(out_l[0]), len(out_r[0])))
    return np.concatenate([out_l, out_r]).T


@hydra.main(config_path=".", config_name="config")
def run_HL_processing(cfg: DictConfig) -> None:
    output_path = os.path.join(cfg["path"]["exp_folder"], "eval_signals")
    os.makedirs(output_path, exist_ok=True)
    scenes_listeners = json.load(open(cfg["path"]["scenes_listeners_file"]))
    listener_audiograms = json.load(open(cfg["path"]["listeners_file"]))
    enhanced_folder = cfg["path"]["enhanced_signals"]

    # initialize ear
    ear = Ear(**cfg["MSBGEar"])

    for scene in tqdm(scenes_listeners):
        for listener in scenes_listeners[scene]:
            if os.path.exists(enhanced_folder):
                signal_file = os.path.join(
                    enhanced_folder, f"{scene}_{listener}_HA-output.wav"
                )
            # if no enhanced signals, use the unprocessed signal for si calculation
            else:
                signal_file = os.path.join(
                    cfg["path"]["scenes_folder"], f"{scene}_mixed_CH0.wav"
                )

            # signals to write
            outfile_stem = f"{output_path}/{scene}_{listener}"
            signal_files_to_write = [
                f"{output_path}/{scene}_flat0dB_HL-output.wav",
                f"{outfile_stem}_HL-output.wav",
                f"{outfile_stem}_HLddf-output.wav",
                f"{outfile_stem}_HL-mixoutput.wav",
            ]
            # if all signals to write exist, pass
            if all([os.path.isfile(f) for f in signal_files_to_write]):
                continue

            signal = read_signal(signal_file)
            mixture_signal = read_signal(
                os.path.join(cfg["path"]["scenes_folder"], f"{scene}_mixed_CH0.wav")
            )

            # retrieve audiograms
            cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
            audiogram_left = np.array(
                listener_audiograms[listener]["audiogram_levels_l"]
            )
            left_audiogram = Audiogram(cfs=cfs, levels=audiogram_left)
            audiogram_right = np.array(
                listener_audiograms[listener]["audiogram_levels_r"]
            )
            right_audiogram = Audiogram(cfs=cfs, levels=audiogram_right)

            # Create discrete delta function (DDF) signal for time alignment
            ddf_signal = np.zeros((np.shape(signal)))
            ddf_signal[:, 0] = unit_impulse(len(signal), int(MSBG_FS / 2))
            ddf_signal[:, 1] = unit_impulse(len(signal), int(MSBG_FS / 2))

            # Get flat-0dB ear audiograms
            flat0dB_audiogram = Audiogram(cfs=cfs, levels=np.zeros((np.shape(cfs))))

            signals_to_write = [
                listen(ear, ddf_signal, flat0dB_audiogram, flat0dB_audiogram),
                listen(ear, signal, left_audiogram, right_audiogram),
                listen(ear, ddf_signal, left_audiogram, right_audiogram),
                listen(ear, mixture_signal, left_audiogram, right_audiogram),
            ]
            for i in range(len(signals_to_write)):
                write_signal(
                    signal_files_to_write[i],
                    signals_to_write[i],
                    MSBG_FS,
                    floating_point=True,
                )


@hydra.main(config_path=".", config_name="config")
def run_calculate_SI(cfg: DictConfig) -> None:
    scenes_listeners = json.load(open(cfg["path"]["scenes_listeners_file"]))
    proc_folder = os.path.join(cfg["path"]["exp_folder"], "eval_signals")
    sii_file = os.path.join(cfg["path"]["exp_folder"], "sii.csv")
    csv_lines = [["scene", "listener", "sii"]]

    for scene in tqdm(scenes_listeners):
        for listener in scenes_listeners[scene]:
            logging.info(f"Running SI calculation: scene {scene}, listener {listener}")
            proc = read_signal(
                os.path.join(proc_folder, f"{scene}_{listener}_HL-output.wav")
            )
            clean = read_signal(
                os.path.join(
                    cfg["path"]["scenes_folder"], f"{scene}_target_anechoic.wav"
                )
            )
            ddf = read_signal(
                os.path.join(proc_folder, f"{scene}_{listener}_HLddf-output.wav")
            )

            # Calculate channel-specific unit impulse delay due to HL model and audiograms
            delay = find_delay_impulse(ddf, initial_value=int(MSBG_FS / 2))
            if delay[0] != delay[1]:
                logging.info(f"Difference in delay of {delay[0] - delay[1]}.")
            maxdelay = int(np.max(delay))

            # Allow for value lower than 1000 samples in case of unimpaired hearing
            if maxdelay > 2000:
                logging.error("Error in delay calculation for signal time-alignment.")

            # Correct for delays by padding clean signals
            cleanpad = np.zeros((len(clean) + maxdelay, 2))
            procpad = np.zeros((len(clean) + maxdelay, 2))

            if len(procpad) < len(proc):
                raise ValueError("Padded processed signal is too short.")

            cleanpad[int(delay[0]) : int(len(clean) + int(delay[0])), 0] = clean[:, 0]
            cleanpad[int(delay[1]) : int(len(clean) + int(delay[1])), 1] = clean[:, 1]
            procpad[: len(proc)] = proc

            sii = mbstoi(
                cleanpad[:, 0],
                cleanpad[:, 1],
                procpad[:, 0],
                procpad[:, 1],
                cfg["mbstoi"]["fs"],
                cfg["mbstoi"]["gridcoarseness"],
            )
            csv_lines.append([scene, listener, sii])  # type: ignore

    with open(sii_file, "w") as csv_f:
        csv_writer = csv.writer(
            csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for line in csv_lines:
            csv_writer.writerow(line)


if __name__ == "__main__":
    run_HL_processing()
    run_calculate_SI()



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

EPS = 1e-8


class AudiometricFIR(nn.Module):
    def __init__(self, sr=44100, nfir=220, device=None):
        super().__init__()
        if device is not None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device
        self.window_size = nfir + 1
        self.padding = nfir // 2

        aud = np.array([250, 500, 1000, 2000, 4000, 6000])
        aud_fv = np.append(np.append(0, aud), sr // 2)  # Audiometric frequency vector
        linear_fv = (
            np.linspace(0, nfir, nfir + 1) / nfir * sr // 2
        )  # linear frequency vector
        interval_freq = np.zeros([len(linear_fv), 2])
        interval_idx = np.zeros([len(linear_fv), 2], dtype=int)
        for i, linear_fv_i in enumerate(linear_fv):
            for j in range(len(aud_fv) - 1):
                if aud_fv[j] <= linear_fv_i < aud_fv[j + 1]:
                    interval_freq[i, 0] = aud_fv[j]
                    interval_freq[i, 1] = aud_fv[j + 1]
                    interval_idx[i, 0] = j
                    interval_idx[i, 1] = j + 1
        interval_freq[-1, 0] = aud_fv[-2]
        interval_freq[-1, 1] = aud_fv[-1]
        interval_idx[-1, 0] = len(aud_fv) - 2
        interval_idx[-1, 1] = len(aud_fv) - 1

        self.interval_idx = interval_idx
        x2_minus_x1 = interval_freq[:, 1] - interval_freq[:, 0]
        x_minus_x1 = linear_fv - interval_freq[:, 0]
        self.x2_minus_x1 = torch.tensor(
            x2_minus_x1, dtype=torch.float32, device=self.device
        )
        self.x_minus_x1 = torch.tensor(
            x_minus_x1, dtype=torch.float32, device=self.device
        )

        self.amp = nn.Parameter(
            torch.tensor(
                np.ones_like(aud_fv[1:-1]), dtype=torch.float32, device=self.device
            )
        )

    def forward(self, x):
        amp = torch.pow(10, torch.abs(self.amp) / 20.0)
        amp = torch.cat((torch.cat((amp[:1], amp)), amp[-1:]))
        y = amp[self.interval_idx]
        y2_minus_y1 = y[:, 1] - y[:, 0]
        y1 = y[:, 0]
        gain = y2_minus_y1 * self.x_minus_x1 / self.x2_minus_x1 + y1

        # firwin
        phase = torch.zeros_like(gain, device=self.device)
        gain = gain.unsqueeze(1)
        phase = phase.unsqueeze(1)
        magnitudes = torch.view_as_complex(torch.cat([gain, phase], dim=1).unsqueeze(0))
        impulse_response = torch.fft.irfft(magnitudes, dim=1)

        window_size = self.window_size + 1
        window = torch.hann_window(window_size, device=self.device)
        ir_size = int(impulse_response.shape[-1])
        half_idx = (window_size + 1) // 2
        padding = ir_size - window_size
        window = torch.cat(
            [
                window[half_idx:],
                torch.zeros([padding], device=self.device),
                window[:half_idx],
            ],
            dim=0,
        )
        impulse_response = impulse_response * window
        first_half_start = ir_size - (half_idx - 1)
        second_half_end = half_idx
        fir_filter = torch.cat(
            [
                impulse_response[:, first_half_start:],
                impulse_response[:, :second_half_end],
            ],
            dim=-1,
        ).unsqueeze(0)

        # x = x.unsqueeze(1)
        output = F.conv1d(x, fir_filter, padding=self.padding, bias=None)
        return output



import json
import logging
import os

import hydra
import speechbrain as sb
import torch
from fastdtw import fastdtw
from hyperpyyaml import load_hyperpyyaml
from omegaconf import DictConfig
from scipy.spatial.distance import cosine
from speechbrain.utils.distributed import run_on_main
from tqdm import tqdm
from transfromer_cpc1_decoder import S2STransformerBeamSearch

logger = logging.getLogger(__name__)


class ASR(sb.core.Brain):
    def generate_feats(self, wavs, wav_lens, tokens_bos):
        """Forward computations from the waveform batches to the output probabilities."""
        # batch = batch.to(self.device)
        wavs, wav_lens, tokens_bos = (
            wavs.to(self.device),
            wav_lens.to(self.device),
            tokens_bos.to(self.device),
        )
        with torch.no_grad():
            feats = self.hparams.compute_features(wavs)
            current_epoch = self.hparams.epoch_counter.current
            feats = self.hparams.normalize(feats, wav_lens, epoch=current_epoch)

            cnn_out = self.hparams.CNN(feats)
            enc_out, _ = self.hparams.Transformer(
                cnn_out, tokens_bos, wav_lens, pad_idx=self.hparams.pad_index
            )
            _, _, dec_out, _ = self.test_search(enc_out.detach(), wav_lens)

        return enc_out.detach().cpu(), dec_out.unsqueeze(0).detach().cpu()

    def init_evaluation(self, max_key=None, min_key=None):
        """perform checkpoint averge if needed"""
        super().on_evaluate_start()

        ckpts = self.checkpointer.find_checkpoints(max_key=max_key, min_key=min_key)
        ckpt = sb.utils.checkpoints.average_checkpoints(
            ckpts, recoverable_name="model", device=self.device
        )
        self.hparams.model.load_state_dict(ckpt, strict=True)
        self.hparams.model.eval()

        self.test_search = S2STransformerBeamSearch(
            modules=[
                self.hparams.Transformer,
                self.hparams.seq_lin,
                self.hparams.ctc_lin,
            ],
            bos_index=self.hparams.bos_index,
            eos_index=self.hparams.eos_index,
            blank_index=self.hparams.blank_index,
            min_decode_ratio=self.hparams.min_decode_ratio,
            max_decode_ratio=self.hparams.max_decode_ratio,
            beam_size=self.hparams.test_beam_size,
            ctc_weight=self.hparams.ctc_weight_decode,
            lm_weight=self.hparams.lm_weight,
            lm_modules=self.hparams.lm_model,
            temperature=1,
            temperature_lm=1,
            topk=10,
            using_eos_threshold=False,
            length_normalization=True,
        )


def init_asr(asr_config):
    hparams_file, run_opts, overrides = sb.parse_arguments([asr_config])
    with open(hparams_file) as fin:
        hparams = load_hyperpyyaml(fin, overrides)

    tokenizer = hparams["tokenizer"]
    bos_index = hparams["bos_index"]

    # We download the pretrained LM from HuggingFace (or elsewhere depending on
    # the path given in the YAML file). The tokenizer is loaded at the same time.
    run_on_main(hparams["pretrainer"].collect_files)
    hparams["pretrainer"].load_collected(device=run_opts["device"])

    asr_brain = ASR(
        modules=hparams["modules"],
        opt_class=hparams["Adam"],
        hparams=hparams,
        run_opts=run_opts,
        checkpointer=hparams["checkpointer"],
    )
    asr_brain.init_evaluation()

    return asr_brain, tokenizer, bos_index


def dtw_similarity(x, y):
    path = fastdtw(
        x.detach().cpu().numpy()[0], y.detach().cpu().numpy()[0], dist=cosine
    )[1]

    x_, y_ = [], []
    for step in range(len(path)):
        x_.append(x[:, path[step][0], :])
        y_.append(y[:, path[step][1], :])
    x_ = torch.stack(x_, dim=1)
    y_ = torch.stack(y_, dim=1)
    return torch.nn.functional.cosine_similarity(x_, y_, dim=-1)


def feat2similarity(
    left_proc_feats, right_proc_feats, left_ref_feats, right_ref_feats, if_dtw=False
):
    if if_dtw:
        ll_sim = dtw_similarity(left_ref_feats, left_proc_feats)
        lr_sim = dtw_similarity(left_ref_feats, right_proc_feats)
        rl_sim = dtw_similarity(right_ref_feats, left_proc_feats)
        rr_sim = dtw_similarity(right_ref_feats, right_proc_feats)
        sim = torch.max(
            torch.stack(
                [
                    torch.mean(ll_sim, dim=-1),
                    torch.mean(lr_sim, dim=-1),
                    torch.mean(rl_sim, dim=-1),
                    torch.mean(rr_sim, dim=-1),
                ],
                dim=-1,
            ),
            dim=-1,
        )[0]
        return sim
    else:
        max_length = torch.max(
            torch.LongTensor(
                [
                    left_proc_feats.shape[1],
                    right_proc_feats.shape[1],
                    left_ref_feats.shape[1],
                    right_ref_feats.shape[1],
                ]
            )
        )
        padded_proc_feats_left = torch.zeros(
            [1, max_length, left_proc_feats.shape[2]], dtype=torch.float32
        )
        padded_proc_feats_right = torch.zeros(
            [1, max_length, right_proc_feats.shape[2]], dtype=torch.float32
        )
        padded_ref_feats_left = torch.zeros(
            [1, max_length, left_ref_feats.shape[2]], dtype=torch.float32
        )
        padded_ref_feats_right = torch.zeros(
            [1, max_length, right_ref_feats.shape[2]], dtype=torch.float32
        )
        padded_proc_feats_left[:, : left_proc_feats.shape[1], :] = left_proc_feats
        padded_proc_feats_right[:, : right_proc_feats.shape[1], :] = right_proc_feats
        padded_ref_feats_left[:, : left_ref_feats.shape[1], :] = left_ref_feats
        padded_ref_feats_right[:, : right_ref_feats.shape[1], :] = right_ref_feats

        ll_sim = torch.nn.functional.cosine_similarity(
            padded_ref_feats_left, padded_proc_feats_left, dim=-1
        )
        lr_sim = torch.nn.functional.cosine_similarity(
            padded_ref_feats_left, padded_proc_feats_right, dim=-1
        )
        rl_sim = torch.nn.functional.cosine_similarity(
            padded_ref_feats_right, padded_proc_feats_left, dim=-1
        )
        rr_sim = torch.nn.functional.cosine_similarity(
            padded_ref_feats_right, padded_proc_feats_right, dim=-1
        )
        sim = torch.stack([ll_sim, lr_sim, rl_sim, rr_sim], dim=-1).max(dim=-1)[0]
        return torch.mean(sim, dim=-1)


def compute_similarity(left_proc_path, wrd, asr_model, bos_index, tokenizer):
    wav_len = torch.tensor([1], dtype=torch.float32)
    tokens_bos = torch.LongTensor([bos_index] + (tokenizer.encode_as_ids(wrd))).view(
        1, -1
    )

    left_ref_path = left_proc_path.replace("msbg", "ref")
    right_proc_path = left_proc_path.replace("left", "right")
    right_ref_path = right_proc_path.replace("msbg", "ref")

    left_proc_wav = sb.dataio.dataio.read_audio(left_proc_path).view(1, -1)
    left_ref_wav = sb.dataio.dataio.read_audio(left_ref_path).view(1, -1)
    right_proc_wav = sb.dataio.dataio.read_audio(right_proc_path).view(1, -1)
    right_ref_wav = sb.dataio.dataio.read_audio(right_ref_path).view(1, -1)

    left_proc_feats = asr_model.generate_feats(left_proc_wav, wav_len, tokens_bos)
    left_ref_feats = asr_model.generate_feats(left_ref_wav, wav_len, tokens_bos)
    right_proc_feats = asr_model.generate_feats(right_proc_wav, wav_len, tokens_bos)
    right_ref_feats = asr_model.generate_feats(right_ref_wav, wav_len, tokens_bos)

    enc_similarity = feat2similarity(
        left_proc_feats[0], right_proc_feats[0], left_ref_feats[0], right_ref_feats[0]
    )
    dec_similarity = feat2similarity(
        left_proc_feats[1],
        right_proc_feats[1],
        left_ref_feats[1],
        right_ref_feats[1],
        if_dtw=True,
    )
    return enc_similarity[0].numpy(), dec_similarity[0].numpy()


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    if cfg.cpc1_track == "open":
        track = "_indep"
    elif cfg.cpc1_track == "closed":
        track = ""
    else:
        logger.error("cpc1_track has to be closed or open")

    asr_model, tokenizer, bos_index = init_asr(cfg.asr_config)

    left_dev_csv = sb.dataio.dataio.load_data_csv(
        os.path.join(cfg.path.exp_folder, "cpc1_asr_data" + track, "left_dev_msbg.csv")
    )  # using left ear csvfile for data loading
    left_test_csv = sb.dataio.dataio.load_data_csv(
        os.path.join(cfg.path.exp_folder, "cpc1_asr_data" + track, "left_test_msbg.csv")
    )  # using left ear csvfile for data loading

    # dev set similarity
    dev_enc_similarity = {}
    dev_dec_similarity = {}
    for wav_id, wav_obj in tqdm(left_dev_csv.items()):
        left_proc_path = wav_obj["wav"]
        wrd = wav_obj["wrd"]
        similarity = compute_similarity(
            left_proc_path, wrd, asr_model, bos_index, tokenizer
        )
        dev_enc_similarity[wav_id] = similarity[0].tolist()
        dev_dec_similarity[wav_id] = similarity[1].tolist()

        with open(
            os.path.join(cfg.path.exp_folder, "dev_enc_similarity.json"), "w"
        ) as f:
            json.dump(dev_enc_similarity, f)
        with open(
            os.path.join(cfg.path.exp_folder, "dev_dec_similarity.json"), "w"
        ) as f:
            json.dump(dev_dec_similarity, f)

    # test set similarity
    test_enc_similarity = {}
    test_dec_similarity = {}
    for wav_id, wav_obj in tqdm(left_test_csv.items()):
        left_proc_path = wav_obj["wav"]
        wrd = wav_obj["wrd"]
        similarity = compute_similarity(
            left_proc_path, wrd, asr_model, bos_index, tokenizer
        )
        test_enc_similarity[wav_id] = similarity[0].tolist()
        test_dec_similarity[wav_id] = similarity[1].tolist()

        with open(
            os.path.join(cfg.path.exp_folder, "test_enc_similarity.json"), "w"
        ) as f:
            json.dump(test_enc_similarity, f)
        with open(
            os.path.join(cfg.path.exp_folder, "test_dec_similarity.json"), "w"
        ) as f:
            json.dump(test_dec_similarity, f)


if __name__ == "__main__":
    run()



import json
import logging
import os

import hydra
import numpy as np
import pytorch_lightning as pl
import torch
import torchaudio
from omegaconf import DictConfig
from pytorch_lightning.callbacks import ModelCheckpoint
from torch.utils.data import DataLoader

from clarity.dataset.cec1_dataset import CEC1Dataset
from clarity.engine.losses import SNRLoss, STOILevelLoss
from clarity.engine.system import System
from clarity.enhancer.dnn.mc_conv_tasnet import ConvTasNet
from clarity.enhancer.dsp.filter import AudiometricFIR
from clarity.predictor.torch_msbg import MSBGHearingModel

logger = logging.getLogger(__name__)


class DenModule(System):
    def common_step(self, batch, batch_nb, train=True):
        proc, ref = batch
        ref = ref[:, self.ear_idx, :]
        if self.config.downsample_factor != 1:
            proc = self.down_sample(proc)
            ref = self.down_sample(ref)
        enhanced = self.model(proc).squeeze(1)
        loss = self.loss_func(enhanced, ref)
        return loss


class AmpModule(System):
    def common_step(self, batch, batch_nb, train=True):
        proc, ref = batch
        ref = ref[:, self.ear_idx, :]
        if self.config.downsample_factor != 1:
            proc = self.down_sample(proc)
            ref = self.down_sample(ref)
        enhanced = self.model(self.den_model(proc)).squeeze(1)

        if self.config.downsample_factor != 1:
            enhanced = torch.clamp(self.up_sample(enhanced), -1, 1)
            ref = torch.clamp(self.up_sample(ref), -1, 1)

        sim_ref = self.nh_ear(ref)
        sim_enhanced = self.hl_ear(enhanced)
        loss = self.loss_func(sim_enhanced, sim_ref)
        return loss


def train_den(cfg, ear):
    exp_dir = os.path.join(cfg.path.exp_folder, ear + "_den")
    if os.path.exists(os.path.join(exp_dir, "best_model.pth")):
        logger.info("Enhancement module exist")
        return

    train_set = CEC1Dataset(**cfg.train_dataset)
    train_loader = DataLoader(dataset=train_set, **cfg.train_loader)
    dev_set = CEC1Dataset(**cfg.dev_dataset)
    dev_loader = DataLoader(dataset=dev_set, **cfg.dev_loader)

    den_model = ConvTasNet(**cfg.mc_conv_tasnet)
    optimizer = torch.optim.Adam(
        params=den_model.parameters(), **cfg.den_trainer.optimizer
    )
    loss_func = SNRLoss()

    den_module = DenModule(
        model=den_model,
        loss_func=loss_func,
        optimizer=optimizer,
        train_loader=train_loader,
        val_loader=dev_loader,
        config=cfg,
    )
    den_module.ear_idx = 0 if ear == "left" else 1
    if cfg.downsample_factor != 1:
        den_module.down_sample = torchaudio.transforms.Resample(
            orig_freq=cfg.sr,
            new_freq=cfg.sr // cfg.downsample_factor,
            resampling_method="sinc_interpolation",
        )

    # callbacks
    callbacks = []
    checkpoint_dir = os.path.join(exp_dir, "checkpoints/")
    checkpoint = ModelCheckpoint(
        checkpoint_dir, monitor="val_loss", mode="min", save_top_k=5, verbose=True
    )
    callbacks.append(checkpoint)

    # set device
    gpus = -1 if torch.cuda.is_available() else None

    trainer = pl.Trainer(
        max_epochs=cfg.den_trainer.epochs,
        callbacks=callbacks,
        default_root_dir=exp_dir,
        gpus=gpus,
        limit_train_batches=1.0,  # Useful for fast experiment
        gradient_clip_val=cfg.den_trainer.gradient_clip_val,
    )
    trainer.fit(den_module)

    best_k = {k: v.item() for k, v in checkpoint.best_k_models.items()}
    with open(os.path.join(exp_dir, "best_k_models.json"), "w") as f:
        json.dump(best_k, f, indent=0)
    state_dict = torch.load(checkpoint.best_model_path)
    den_module.load_state_dict(state_dict=state_dict["state_dict"])
    den_module.cpu()
    torch.save(den_module.model.state_dict(), os.path.join(exp_dir, "best_model.pth"))


def train_amp(cfg, ear):
    exp_dir = os.path.join(cfg.path.exp_folder, ear + "_amp")
    if os.path.exists(os.path.join(exp_dir, "best_model.pth")):
        logger.info("Amplification module exist")
        return

    train_set = CEC1Dataset(**cfg.train_dataset)
    train_loader = DataLoader(dataset=train_set, **cfg.train_loader)
    dev_set = CEC1Dataset(**cfg.dev_dataset)
    dev_loader = DataLoader(dataset=dev_set, **cfg.dev_loader)

    # load denoising module
    den_model = ConvTasNet(**cfg.mc_conv_tasnet)
    den_model_path = os.path.join(
        os.path.join(cfg.path.exp_folder, ear + "_den"), "best_model.pth"
    )
    den_model.load_state_dict(torch.load(den_model_path))

    # amplification module
    amp_model = AudiometricFIR(**cfg.fir)
    optimizer = torch.optim.Adam(
        params=amp_model.parameters(), **cfg.amp_trainer.optimizer
    )
    loss_func = STOILevelLoss(**cfg.amp_trainer.stoilevel_loss)

    amp_module = AmpModule(
        model=amp_model,
        loss_func=loss_func,
        optimizer=optimizer,
        train_loader=train_loader,
        val_loader=dev_loader,
        config=cfg,
    )
    amp_module.ear_idx = 0 if ear == "left" else 1
    amp_module.den_model = den_model
    if cfg.downsample_factor != 1:
        amp_module.down_sample = torchaudio.transforms.Resample(
            orig_freq=cfg.sr,
            new_freq=cfg.sr // cfg.downsample_factor,
            resampling_method="sinc_interpolation",
        )
        amp_module.up_sample = torchaudio.transforms.Resample(
            orig_freq=cfg.sr // cfg.downsample_factor,
            new_freq=cfg.sr,
            resampling_method="sinc_interpolation",
        )

    # build normal hearing and hearing loss ears
    with open(cfg.listener.metafile, "r") as f:
        listeners_file = json.load(f)
        audiogram_cfs = listeners_file[cfg.listener.id]["audiogram_cfs"]
        audiogram_lvl_l = listeners_file[cfg.listener.id]["audiogram_levels_l"]
        audiogram_lvl_r = listeners_file[cfg.listener.id]["audiogram_levels_r"]
    audiogram = audiogram_lvl_l if ear == "left" else audiogram_lvl_r

    amp_module.nh_ear = MSBGHearingModel(
        audiogram=np.zeros_like(audiogram), audiometric=audiogram_cfs, sr=cfg.sr
    )
    amp_module.hl_ear = MSBGHearingModel(
        audiogram=audiogram, audiometric=audiogram_cfs, sr=cfg.sr
    )

    # callbacks
    callbacks = []
    checkpoint_dir = os.path.join(exp_dir, "checkpoints/")
    checkpoint = ModelCheckpoint(
        checkpoint_dir, monitor="val_loss", mode="min", save_top_k=5, verbose=True
    )
    callbacks.append(checkpoint)

    # set device
    gpus = -1 if torch.cuda.is_available() else None

    trainer = pl.Trainer(
        max_epochs=cfg.amp_trainer.epochs,
        callbacks=callbacks,
        default_root_dir=exp_dir,
        gpus=gpus,
        limit_train_batches=1.0,  # Useful for fast experiment
        gradient_clip_val=cfg.amp_trainer.gradient_clip_val,
    )
    trainer.fit(amp_module)

    best_k = {k: v.item() for k, v in checkpoint.best_k_models.items()}
    with open(os.path.join(exp_dir, "best_k_models.json"), "w") as f:
        json.dump(best_k, f, indent=0)
    state_dict = torch.load(checkpoint.best_model_path)
    amp_module.load_state_dict(state_dict=state_dict["state_dict"])
    amp_module.cpu()
    torch.save(amp_module.model.state_dict(), os.path.join(exp_dir, "best_model.pth"))


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    logger.info("Begin training left ear enhancement module.")
    train_den(cfg, ear="left")
    logger.info("Begin training right ear enhancement module.")
    train_den(cfg, ear="right")
    logger.info("Begin training left ear amplification module.")
    train_amp(cfg, ear="left")
    logger.info("Begin training right ear amplification module.")
    train_amp(cfg, ear="right")


if __name__ == "__main__":
    run()



from clarity.evaluator.haspi import eb, ebm, ip


def haspi_v2(x, fx, y, fy, HL, Level1=65):
    """
    Compute the HASPI intelligibility index using the
    auditory model followed by computing the envelope cepstral
    correlation and BM vibration high-level covariance. The reference
    signal presentation level for NH listeners is assumed to be 65 dB
    SPL. The same model is used for both normal and impaired hearing. This
    version of HASPI uses a modulation filterbank followed by an ensemble of
    neural networks to compute the estimated intelligibility.

    Args:
        x (): Clear input reference speech signal with no noise or distortion.
              If a hearing loss is specified, no amplification should be provided.
        fx (): Sampling rate in Hz for signal x
        y (): Output signal with noise, distortion, HA gain, and/or processing.
        fy (): Sampling rate in Hz for signal y.
        HL (): (1,6) vector of hearing loss at the 6 audiometric frequencies
                    [250, 500, 1000, 2000, 4000, 6000] Hz.
        Level1 (): Optional input specifying level in dB SPL that corresponds to a
              signal RMS = 1. Default is 65 dB SPL if argument not provided.

    Returns:
    Intel     Intelligibility estimated by passing the cepstral coefficients
              through a modulation filterbank followed by an ensemble of
              neural networks.
    raw       vector of 10 cep corr modulation filterbank outputs, averaged
              over basis funct 2-6.

    James M. Kates, 5 August 2013.

    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Auditory model for intelligibility
    # Reference is no processing, normal hearing
    itype = 0  # Intelligibility model
    xenv, _, yenv, _, _, _, fsamp = eb.EarModel(x, fx, y, fy, HL, itype, Level1)

    # Envelope modulation features

    # LP filter and subsample the envelope
    fLP = 320
    fsub = 8 * fLP  # subsample to span 2 octaves above the cutoff frequency
    xLP, yLP = ebm.EnvFilt(xenv, yenv, fLP, fsub, fsamp)

    # Compute the ceptstral coefficients as a function of subsampled time
    nbasis = 6  # Use 6 basis functions
    thr = 2.5  # Silence threshold in dB SL
    dither = 0.1  # Dither in dB RMS to add to envelope signals
    xcep, ycep = ebm.CepCoef(xLP, yLP, thr, dither, nbasis)

    # Cepstral coeffifiencts filtered at each modulation rate
    # Band center frequencies [2, 6, 10, 16, 25, 40, 64, 100, 160, 256] Hz
    # Band edges [0, 4, 8, 12.5, 20.5, 30.5, 52.4, 78.1, 128, 200, 328] Hz
    xmod, ymod, _ = ebm.ModFilt(xcep, ycep, fsub)

    # Cross-correlation between the cepstral coefficients for the degraded and
    # ref signals at each modulation rate, averaged over basis functions 2-6
    aveCM = ebm.ModCorr(xmod, ymod)

    # Intelligibility prediction
    # Get the neural network parameters and the weights for an ensemble of 10 networks
    NNparam, Whid, Wout, b = ip.GetNeuralNet()

    # Average the neural network outputs for the modulation filterbank values
    model = ip.NNfeedfwdEns(aveCM, NNparam, Whid, Wout)
    model = model[0] / b

    # Return the intelligibility estimate and raw modulation filter outputs
    Intel = model
    raw = aveCM

    return Intel, raw


def haspi_v2_be(
    xl, xr, yl, yr, fs_signal, audiogram_l, audiogram_r, audiogram_cfs, Level=100
) -> float:
    """Better ear HASPI.

    Calculates HASPI for left and right ear and selects the better result.

    Args:
        xl: left channel of reference signal
        xr: right channel of reference signal
        yl: left channel of processed signal
        yr: right channel of processed signal
        fs_signal: sampling rate for both signal
        audiogram_l: left ear audiogram
        audiogram_r: right ear audiogram
        audiogram_cfs: audiogram frequencies
        Level: level in dB SPL corresponding to RMS=1

    Returns:
        float: beHASPI score

    Zuzanna Podwinska, March 2022
    """

    # HASPI assumes the following audiogram frequencies:
    aud = [250, 500, 1000, 2000, 4000, 6000]

    # Adjust listener.audiogram_levels_l and _r to match the frequencies above
    HL_l = [
        audiogram_l[i] for i in range(len(audiogram_cfs)) if audiogram_cfs[i] in aud
    ]
    HL_r = [
        audiogram_r[i] for i in range(len(audiogram_cfs)) if audiogram_cfs[i] in aud
    ]

    score_l, _ = haspi_v2(xl, fs_signal, yl, fs_signal, HL_l, Level)
    score_r, _ = haspi_v2(xr, fs_signal, yr, fs_signal, HL_r, Level)

    return max(score_l, score_r)



import os
import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from typing import Union

import numpy as np
import scipy
import scipy.io

SPEECH_FILTER = scipy.io.loadmat(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "params", "speech_weight.mat"
    ),
    squeeze_me=True,
)
SPEECH_FILTER = np.array(SPEECH_FILTER["filt"])


def better_ear_speechweighted_snr(target: np.ndarray, noise: np.ndarray) -> float:
    """Calculate effective better ear SNR.

    Args:
        target (np.ndarray):
        noise (np.ndarray):

    Returns:
    """
    if np.ndim(target) == 1:
        # analysis left ear and right ear for single channel target
        try:
            left_snr = speechweighted_snr(target, noise[:, 0])
            right_snr = speechweighted_snr(target, noise[:, 1])
        except IndexError:
            raise
    else:
        # analysis left ear and right ear for two channel target
        left_snr = speechweighted_snr(target[:, 0], noise[:, 0])
        right_snr = speechweighted_snr(target[:, 1], noise[:, 1])
    # snr is the max of left and right
    be_snr = max(left_snr, right_snr)
    return be_snr


def speechweighted_snr(target: np.ndarray, noise: np.ndarray) -> float:
    """Apply speech weighting filter to signals and get SNR.

    Args:
        target (np.ndarray):
        noise (np.ndarray):

    Returns:
        (float):
    Signal Noise Ratio
    """
    try:
        target_filt = scipy.signal.convolve(
            target, SPEECH_FILTER, mode="full", method="fft"
        )
        noise_filt = scipy.signal.convolve(
            noise, SPEECH_FILTER, mode="full", method="fft"
        )
    except ValueError:
        raise
    # rms of the target after speech weighted filter
    targ_rms = np.sqrt(np.mean(target_filt**2))

    # rms of the noise after speech weighted filter
    noise_rms = np.sqrt(np.mean(noise_filt**2))
    sw_snr = np.divide(targ_rms, noise_rms)
    return sw_snr


def sum_signals(signals: list) -> Union[np.ndarray, Literal[0]]:
    """Return sum of a list of signals.

    Signals are stored as a list of ndarrays whose size can vary in the first
    dimension, i.e., so can sum mono or stereo signals etc.
    Shorter signals are zero padded to the length of the longest.

    Args:
        signals (list): List of signals stored as ndarrays

    Returns:
        np.ndarray: The sum of the signals

    """
    max_length = max(x.shape[0] for x in signals)
    return sum(pad(x, max_length) for x in signals)


def pad(signal: np.ndarray, length: int) -> np.ndarray:
    """Zero pad signal to required length.

    Assumes required length is not less than input length.

    Args:
        signal (np.array):
        length (int):

    Returns:
        np.array:
    """
    # FIXME : Consider encapsulating in a 'try: ... except: ...' should the assertion not be met.
    try:
        assert length >= signal.shape[0]
    except AssertionError:
        raise
    return np.pad(
        signal, ([(0, length - signal.shape[0])] + ([(0, 0)] * (len(signal.shape) - 1)))
    )



import json
import logging
import os

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.optimize import curve_fit
from scipy.stats import kendalltau, pearsonr

logger = logging.getLogger(__name__)


def rmse_score(x, y):
    return np.sqrt(np.mean((x - y) ** 2))


def ncc_score(x, y):
    return pearsonr(x, y)[0]


def kt_score(x, y):
    return kendalltau(x, y)[0]


def std_err(x, y):
    return np.std(x - y) / np.sqrt(len(x))


class Model:
    """Class to represent the mapping from mbstoi parameters to intelligibility scores.
    The mapping uses a simple logistic function scaled between 0 and 100.
    The mapping parameters need to fit first using mbstoi, intelligibility score pairs, using fit().
    Once the fit has been made predictions can be made by calling predict()
    """

    params = None  # The model params

    def _logistic_mapping(self, x, x0, k):
        """
        Logistic function
            x0 - x value of the logistic's midpoint
            k - the logistic growth rate or steepness of the curve
        """
        L = 100  # correctness can't be over 100
        return L / (1 + np.exp(-k * (x - x0)))

    def fit(self, pred, intel):
        """Fit a mapping betweeen mbstoi scores and intelligibility scores."""
        initial_guess = [50.0, 1.0]  # Initial guess for parameter values
        self.params, pcov = curve_fit(
            self._logistic_mapping, pred, intel, initial_guess
        )

    def predict(self, x):
        """Predict intelligilbity scores from mbstoi scores."""
        # Note, fit() must be called before predictions can be made
        assert self.params is not None
        return self._logistic_mapping(x, self.params[0], self.params[1])


def compute_scores(predictions, labels):
    return {
        "RMSE": rmse_score(predictions, labels),
        "Std": std_err(predictions, labels),
        "NCC": ncc_score(predictions, labels),
        "KT": kt_score(predictions, labels),
    }


def read_data(pred_json, label_json):
    prediction = []
    label = []

    # read label_json to dict
    label_dict = {}
    label_json = json.load(open(label_json))
    for item in label_json:
        label_dict[item["signal"]] = item["correctness"]

    pred_dict = json.load(open(pred_json))
    for signal, pred in pred_dict.items():
        prediction.append(pred * 100.0)
        label.append(label_dict[signal])

    return np.array(prediction), np.array(label)


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    if cfg.cpc1_track == "open":
        track = "_indep"
    elif cfg.cpc1_track == "closed":
        track = ""
    else:
        logger.error("cpc1_track has to be closed or open")

    # encoder representation evaluation
    prediction_dev, label_dev = read_data(
        os.path.join(cfg.path.exp_folder, "dev_enc_similarity.json"),
        os.path.join(cfg.path.cpc1_train_data, f"metadata/CPC1.{'train'+track}.json"),
    )
    prediction_test, label_test = read_data(
        os.path.join(cfg.path.exp_folder, "test_enc_similarity.json"),
        f"../test_listener_responses/CPC1.{'test'+track}.json",
    )

    logger.info("Apply logistic fitting.")
    model = Model()
    model.fit(prediction_dev, label_dev)
    fit_pred = model.predict(prediction_test)
    enc_scores = compute_scores(fit_pred, label_test)

    # decoder representation evaluation
    prediction_dev, label_dev = read_data(
        os.path.join(cfg.path.exp_folder, "dev_dec_similarity.json"),
        os.path.join(cfg.path.cpc1_train_data, f"metadata/CPC1.{'train'+track}.json"),
    )
    prediction_test, label_test = read_data(
        os.path.join(cfg.path.exp_folder, "test_dec_similarity.json"),
        f"../test_listener_responses/CPC1.{'test'+track}.json",
    )

    logger.info("Apply logistic fitting.")
    model = Model()
    model.fit(prediction_dev, label_dev)
    fit_pred = model.predict(prediction_test)
    dec_scores = compute_scores(fit_pred, label_test)

    with open(os.path.join(cfg.path.exp_folder, "results.json"), "w") as f:
        json.dump({"enc_results": enc_scores, "dec_results": dec_scores}, f)


if __name__ == "__main__":
    run()



import json
import logging
import os

import hydra
from omegaconf import DictConfig
from tqdm import tqdm

from clarity.data.scene_renderer_cec1 import Renderer, check_scene_exists


def prepare_data(root_path, metafile_path, scene_folder, num_channels):
    """
    Generate scene data given dataset (train or dev)
    Args:
        root_path: Clarity root path
        metafile_path: scene metafile path
        scene_folder: folder containing generated scenes
        num_channels: number of channels
    """
    with open(metafile_path, "r") as f:
        scenes = json.load(f)

    os.makedirs(scene_folder, exist_ok=True)

    renderer = Renderer(input_path=root_path, output_path=scene_folder, num_channels=3)
    for scene in tqdm(scenes):
        if check_scene_exists(scene, scene_folder, num_channels):
            logging.info(f"Skipping processed scene {scene['scene']}.")
        else:
            renderer.render(
                pre_samples=scene["pre_samples"],
                post_samples=scene["post_samples"],
                dataset=scene["dataset"],
                target=scene["target"]["name"],
                noise_type=scene["interferer"]["type"],
                interferer=scene["interferer"]["name"],
                room=scene["room"]["name"],
                scene=scene["scene"],
                offset=scene["interferer"]["offset"],
                snr_dB=scene["SNR"],
            )


@hydra.main(config_path=".", config_name="data_config")
def run(cfg: DictConfig) -> None:
    for dataset in cfg["datasets"]:
        prepare_data(
            cfg["input_path"],
            cfg["datasets"][dataset]["metafile_path"],
            cfg["datasets"][dataset]["scene_folder"],
            cfg["num_channels"],
        )


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
"""
Recipe for training a Transformer ASR system with librispeech, from the SpeechBrain LibriSpeech/ASR recipe.
The SpeechBrain version used in this work is:
https://github.com/speechbrain/speechbrain/tree/1eddf66eea01866d3cf9dfe61b00bb48d2062236
"""

import logging
import os
import sys
from pathlib import Path

import speechbrain as sb
import torch
from hyperpyyaml import load_hyperpyyaml
from speechbrain.utils.distributed import run_on_main

logger = logging.getLogger(__name__)


# Define training procedure
class ASR(sb.core.Brain):
    def compute_forward(self, batch, stage):
        """Forward computations from the waveform batches to the output probabilities."""
        batch = batch.to(self.device)
        wavs, wav_lens = batch.sig
        tokens_bos, _ = batch.tokens_bos

        # compute features
        feats = self.hparams.compute_features(wavs)
        current_epoch = self.hparams.epoch_counter.current
        feats = self.hparams.normalize(feats, wav_lens, epoch=current_epoch)

        if stage == sb.Stage.TRAIN:
            if hasattr(self.hparams, "augmentation"):
                feats = self.hparams.augmentation(feats)

        # forward modules
        src = self.hparams.CNN(feats)
        enc_out, pred = self.hparams.Transformer(
            src, tokens_bos, wav_lens, pad_idx=self.hparams.pad_index
        )

        # output layer for ctc log-probabilities
        logits = self.hparams.ctc_lin(enc_out)
        p_ctc = self.hparams.log_softmax(logits)

        # output layer for seq2seq log-probabilities
        pred = self.hparams.seq_lin(pred)
        p_seq = self.hparams.log_softmax(pred)

        # Compute outputs
        hyps = None
        if stage == sb.Stage.TRAIN:
            hyps = None
        elif stage == sb.Stage.VALID:
            hyps = None
            current_epoch = self.hparams.epoch_counter.current
            if current_epoch % self.hparams.valid_search_interval == 0:
                # for the sake of efficiency, we only perform beamsearch with limited capacity
                # and no LM to give user some idea of how the AM is doing
                hyps, _ = self.hparams.valid_search(enc_out.detach(), wav_lens)
        elif stage == sb.Stage.TEST:
            hyps, _ = self.hparams.test_search(enc_out.detach(), wav_lens)

        return p_ctc, p_seq, wav_lens, hyps

    def compute_objectives(self, predictions, batch, stage):
        """Computes the loss (CTC+NLL) given predictions and targets."""

        (p_ctc, p_seq, wav_lens, hyps) = predictions

        ids = batch.id
        tokens_eos, tokens_eos_lens = batch.tokens_eos
        tokens, tokens_lens = batch.tokens

        loss_seq = self.hparams.seq_cost(p_seq, tokens_eos, length=tokens_eos_lens)
        loss_ctc = self.hparams.ctc_cost(p_ctc, tokens, wav_lens, tokens_lens)
        loss = (
            self.hparams.ctc_weight * loss_ctc
            + (1 - self.hparams.ctc_weight) * loss_seq
        )

        if stage != sb.Stage.TRAIN:
            current_epoch = self.hparams.epoch_counter.current
            valid_search_interval = self.hparams.valid_search_interval
            if current_epoch % valid_search_interval == 0 or (stage == sb.Stage.TEST):
                # Decode token terms to words
                predicted_words = [
                    tokenizer.decode_ids(utt_seq).split(" ") for utt_seq in hyps
                ]
                target_words = [wrd.split(" ") for wrd in batch.wrd]
                self.wer_metric.append(ids, predicted_words, target_words)

            # compute the accuracy of the one-step-forward prediction
            self.acc_metric.append(p_seq, tokens_eos, tokens_eos_lens)
        return loss

    def fit_batch(self, batch):
        """Train the parameters given a single batch in input"""
        # check if we need to switch optimizer
        # if so change the optimizer from Adam to SGD
        self.check_and_reset_optimizer()

        predictions = self.compute_forward(batch, sb.Stage.TRAIN)
        loss = self.compute_objectives(predictions, batch, sb.Stage.TRAIN)

        # normalize the loss by gradient_accumulation step
        (loss / self.hparams.gradient_accumulation).backward()

        if self.step % self.hparams.gradient_accumulation == 0:
            # gradient clipping & early stop if loss is not fini
            self.check_gradients(loss)

            self.optimizer.step()
            self.optimizer.zero_grad()

            # anneal lr every update
            self.hparams.noam_annealing(self.optimizer)

        return loss.detach()

    def evaluate_batch(self, batch, stage):
        """Computations needed for validation/test batches"""
        with torch.no_grad():
            predictions = self.compute_forward(batch, stage=stage)
            loss = self.compute_objectives(predictions, batch, stage=stage)
        return loss.detach()

    def on_stage_start(self, stage, epoch):
        """Gets called at the beginning of each epoch"""
        if stage != sb.Stage.TRAIN:
            self.acc_metric = self.hparams.acc_computer()
            self.wer_metric = self.hparams.error_rate_computer()

    def on_stage_end(self, stage, stage_loss, epoch):
        """Gets called at the end of a epoch."""
        # Compute/store important stats
        stage_stats = {"loss": stage_loss}
        if stage == sb.Stage.TRAIN:
            self.train_stats = stage_stats
        else:
            stage_stats["ACC"] = self.acc_metric.summarize()
            current_epoch = self.hparams.epoch_counter.current
            valid_search_interval = self.hparams.valid_search_interval
            if current_epoch % valid_search_interval == 0 or stage == sb.Stage.TEST:
                stage_stats["WER"] = self.wer_metric.summarize("error_rate")

        # log stats and save checkpoint at end-of-epoch
        if stage == sb.Stage.VALID and sb.utils.distributed.if_main_process():

            # report different epoch stages according current stage
            current_epoch = self.hparams.epoch_counter.current
            if current_epoch <= self.hparams.stage_one_epochs:
                lr = self.hparams.noam_annealing.current_lr
                steps = self.hparams.noam_annealing.n_steps
                optimizer = self.optimizer.__class__.__name__
            else:
                lr = self.hparams.lr_sgd
                steps = -1
                optimizer = self.optimizer.__class__.__name__

            epoch_stats = {
                "epoch": epoch,
                "lr": lr,
                "steps": steps,
                "optimizer": optimizer,
            }
            self.hparams.train_logger.log_stats(
                stats_meta=epoch_stats,
                train_stats=self.train_stats,
                valid_stats=stage_stats,
            )
            self.checkpointer.save_and_keep_only(
                meta={"ACC": stage_stats["ACC"], "epoch": epoch},
                max_keys=["ACC"],
                num_to_keep=1,
            )

        elif stage == sb.Stage.TEST:
            self.hparams.train_logger.log_stats(
                stats_meta={"Epoch loaded": self.hparams.epoch_counter.current},
                test_stats=stage_stats,
            )
            with open(self.hparams.wer_file, "w") as w:
                self.wer_metric.write_stats(w)

            # save the averaged checkpoint at the end of the evaluation stage
            # delete the rest of the intermediate checkpoints
            # ACC is set to 1.1 so checkpointer only keeps the averaged checkpoint
            self.checkpointer.save_and_keep_only(
                meta={"ACC": 1.1, "epoch": epoch}, max_keys=["ACC"], num_to_keep=1
            )

    def check_and_reset_optimizer(self):
        """reset the optimizer if training enters stage 2"""
        current_epoch = self.hparams.epoch_counter.current
        if not hasattr(self, "switched"):
            self.switched = False
            if isinstance(self.optimizer, torch.optim.SGD):
                self.switched = True

        if self.switched is True:
            return

        if current_epoch > self.hparams.stage_one_epochs:
            self.optimizer = self.hparams.SGD(self.modules.parameters())

            if self.checkpointer is not None:
                self.checkpointer.add_recoverable("optimizer", self.optimizer)

            self.switched = True

    def on_fit_start(self):
        """Initialize the right optimizer on the training start"""
        super().on_fit_start()

        # if the model is resumed from stage two, reinitialize the optimizer
        current_epoch = self.hparams.epoch_counter.current
        current_optimizer = self.optimizer
        if current_epoch > self.hparams.stage_one_epochs:
            del self.optimizer
            self.optimizer = self.hparams.SGD(self.modules.parameters())

            # Load latest checkpoint to resume training if interrupted
            if self.checkpointer is not None:

                # do not reload the weights if training is interrupted right before stage 2
                group = current_optimizer.param_groups[0]
                if "momentum" not in group:
                    return

                self.checkpointer.recover_if_possible(device=torch.device(self.device))

    def on_evaluate_start(self, max_key=None, min_key=None):
        """perform checkpoint averge if needed"""
        super().on_evaluate_start()

        ckpts = self.checkpointer.find_checkpoints(max_key=max_key, min_key=min_key)
        ckpt = sb.utils.checkpoints.average_checkpoints(
            ckpts, recoverable_name="model", device=self.device
        )

        self.hparams.model.load_state_dict(ckpt, strict=True)
        self.hparams.model.eval()


def dataio_prepare(hparams):
    """This function prepares the datasets to be used in the brain class.
    It also defines the data processing pipeline through user-defined functions."""
    data_folder = hparams["data_folder"]

    train_data = sb.dataio.dataset.DynamicItemDataset.from_csv(
        csv_path=hparams["train_csv"], replacements={"data_root": data_folder}
    )

    valid_data = sb.dataio.dataset.DynamicItemDataset.from_csv(
        csv_path=hparams["valid_csv"], replacements={"data_root": data_folder}
    )
    valid_data = valid_data.filtered_sorted(sort_key="duration")

    # test is separate
    test_datasets = {}
    for csv_file in hparams["test_csv"]:
        name = Path(csv_file).stem
        test_datasets[name] = sb.dataio.dataset.DynamicItemDataset.from_csv(
            csv_path=csv_file, replacements={"data_root": data_folder}
        )
        test_datasets[name] = test_datasets[name].filtered_sorted(sort_key="duration")

    datasets = [train_data, valid_data] + [i for k, i in test_datasets.items()]

    # We get the tokenizer as we need it to encode the labels when creating
    # mini-batches.
    tokenizer = hparams["tokenizer"]

    # 2. Define audio pipeline:
    @sb.utils.data_pipeline.takes("wav")
    @sb.utils.data_pipeline.provides("sig")
    def audio_pipeline(wav):
        sig = sb.dataio.dataio.read_audio(wav)
        return sig

    sb.dataio.dataset.add_dynamic_item(datasets, audio_pipeline)

    # 3. Define text pipeline:
    @sb.utils.data_pipeline.takes("wrd")
    @sb.utils.data_pipeline.provides(
        "wrd", "tokens_list", "tokens_bos", "tokens_eos", "tokens"
    )
    def text_pipeline(wrd):
        yield wrd
        tokens_list = tokenizer.encode_as_ids(wrd)
        yield tokens_list
        tokens_bos = torch.LongTensor([hparams["bos_index"]] + (tokens_list))
        yield tokens_bos
        tokens_eos = torch.LongTensor(tokens_list + [hparams["eos_index"]])
        yield tokens_eos
        tokens = torch.LongTensor(tokens_list)
        yield tokens

    sb.dataio.dataset.add_dynamic_item(datasets, text_pipeline)

    # 4. Set output:
    sb.dataio.dataset.set_output_keys(
        datasets, ["id", "sig", "wrd", "tokens_bos", "tokens_eos", "tokens"]
    )
    return train_data, valid_data, test_datasets, tokenizer


if __name__ == "__main__":
    # CLI:
    hparams_file, run_opts, overrides = sb.parse_arguments(sys.argv[1:])
    with open(hparams_file) as fin:
        hparams = load_hyperpyyaml(fin, overrides)

    # If distributed_launch=True then
    # create ddp_group with the right communication protocol
    sb.utils.distributed.ddp_init_group(run_opts)

    # Create experiment directory
    sb.create_experiment_directory(
        experiment_directory=hparams["output_folder"],
        hyperparams_to_save=hparams_file,
        overrides=overrides,
    )

    # here we create the datasets objects as well as tokenization and encoding
    train_data, valid_data, test_datasets, tokenizer = dataio_prepare(hparams)

    # We download the pretrained LM from HuggingFace (or elsewhere depending on
    # the path given in the YAML file). The tokenizer is loaded at the same time.
    run_on_main(hparams["pretrainer"].collect_files)
    hparams["pretrainer"].load_collected(device=run_opts["device"])

    # Trainer initialization
    asr_brain = ASR(
        modules=hparams["modules"],
        opt_class=hparams["Adam"],
        hparams=hparams,
        run_opts=run_opts,
        checkpointer=hparams["checkpointer"],
    )

    # adding objects to trainer:
    asr_brain.tokenizer = hparams["tokenizer"]

    # Training
    asr_brain.fit(
        asr_brain.hparams.epoch_counter,
        train_data,
        valid_data,
        train_loader_kwargs=hparams["train_dataloader_opts"],
        valid_loader_kwargs=hparams["valid_dataloader_opts"],
    )

    # Testing
    for k in test_datasets.keys():  # keys are test_clean, test_other etc
        asr_brain.hparams.wer_file = os.path.join(
            hparams["output_folder"], "wer_{}.txt".format(k)
        )
        asr_brain.evaluate(
            test_datasets[k],
            max_key="ACC",
            test_loader_kwargs=hparams["test_dataloader_opts"],
        )



import os

import hydra
import numpy as np
import torch
import torchaudio
from omegaconf import DictConfig
from soundfile import write
from torch.utils.data import DataLoader
from tqdm import tqdm

from clarity.dataset.cec1_dataset import CEC1Dataset
from clarity.enhancer.dnn.mc_conv_tasnet import ConvTasNet
from clarity.enhancer.dsp.filter import AudiometricFIR


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    output_folder = os.path.join(cfg.path.exp_folder, "enhanced_" + cfg.listener.id)
    os.makedirs(output_folder, exist_ok=True)

    test_set = CEC1Dataset(**cfg.test_dataset)
    test_loader = DataLoader(dataset=test_set, **cfg.test_loader)

    if cfg.downsample_factor != 1:
        down_sample = torchaudio.transforms.Resample(
            orig_freq=cfg.sr,
            new_freq=cfg.sr // cfg.downsample_factor,
            resampling_method="sinc_interpolation",
        )
        up_sample = torchaudio.transforms.Resample(
            orig_freq=cfg.sr // cfg.downsample_factor,
            new_freq=cfg.sr,
            resampling_method="sinc_interpolation",
        )

    device = "cuda" if torch.cuda.is_available() else None

    with torch.no_grad():
        for batch in tqdm(test_loader, desc="testing"):
            noisy, scene = batch
            out = []
            for ear in ["left", "right"]:
                torch.cuda.empty_cache()
                # load denoising module
                den_model = ConvTasNet(**cfg.mc_conv_tasnet)
                den_model_path = os.path.join(
                    os.path.join(cfg.path.exp_folder, ear + "_den"), "best_model.pth"
                )
                den_model.load_state_dict(
                    torch.load(den_model_path, map_location=device)
                )
                _den_model = torch.nn.parallel.DataParallel(den_model.to(device))
                _den_model.eval()

                # load amplification module
                amp_model = AudiometricFIR(**cfg.fir)
                amp_model_path = os.path.join(
                    os.path.join(cfg.path.exp_folder, ear + "_amp"), "best_model.pth"
                )
                amp_model.load_state_dict(
                    torch.load(amp_model_path, map_location=device)
                )
                _amp_model = torch.nn.parallel.DataParallel(amp_model.to(device))
                _amp_model.eval()

                noisy = noisy.to(device)
                if cfg.downsample_factor != 1:
                    proc = down_sample(noisy)
                enhanced = amp_model(den_model(proc)).squeeze(1)
                if cfg.downsample_factor != 1:
                    enhanced = up_sample(enhanced)
                enhanced = torch.clamp(enhanced, -1, 1)
                out.append(enhanced.detach().cpu().numpy()[0])

            out = np.stack(out, axis=0).transpose()
            write(
                os.path.join(
                    output_folder,
                    scene[0] + "_" + cfg.listener.id + "_" + "HA-output.wav",
                ),
                out,
                cfg.sr,
            )


if __name__ == "__main__":
    run()



"""Tools to support higher order ambisonic processing."""
import logging
from typing import List

import numpy as np
from numba import njit
from numba.typed import List as TypedList
from scipy.signal import convolve
from scipy.spatial.transform import Rotation as R
from scipy.special import comb

logger = logging.getLogger(__name__)


# Code for generation ambisonic rotation matrices


@njit
def compute_rotation_matrix(n: int, foa_rotmat: np.ndarray) -> np.ndarray:
    """Generate a rotation matrix to rotate HOA soundfield.
    Based on 'Rotation Matrices for Real Spherical Harmonics. Direct Determination
    by Recursion' Joseph Ivanic and Klaus Ruedenberg J. Phys. Chem. 1996, 100, 15,
    6342–6347 and Gorzel, M., Allen, A., Kelly, I., Kammerl, J., Gungormusler,
    A., Yeh, H., and Boland, F., "Efficient Encoding and Decoding of Binaural Sound
    with Resonance Audio", In proc. of the AES International Conference on Immersive
    and Interactive Audio, March 2019
    Operates on HOA of a given order rotates by azimuth theta
    and elevation phi

    Args:
        order (int): order of ambisonic soundfield
        foa_rotmat (arraylike): rotation matrix to expand

    Returns:
        np.ndarray: HOA rotation matrix
    """
    m = (n + 1) ** 2
    # construct rotamat for given order
    n_vector = np.arange(m)
    rot_mat = np.eye(m)
    rot_mat[1:4, 1:4] = foa_rotmat
    upper_index = np.cumsum(2 * n_vector + 1)
    lower_index = np.roll(upper_index, 1)
    lower_index[0] = 0

    sub_matrices = [np.eye(i * 2 + 1) for i in np.arange(n + 1)]
    sub_matrices[1] = foa_rotmat

    typed_sub_matrices = TypedList()
    [typed_sub_matrices.append(x) for x in sub_matrices]

    if n > 1:
        for i in np.arange(2, n + 1):
            rot_mat, typed_sub_matrices = compute_band_rotation(
                i, typed_sub_matrices, rot_mat
            )

    return rot_mat


@njit
def centred_element(r: np.ndarray, i: int, j: int):
    """Get value from centered element indexing.

    Args:
        r (matrix): input matrix
        i (int): row index
        j (int): column index

    Returns:
        Any: matrix element
    """
    offset = int((r.shape[0] - 1) / 2)
    if i > r.shape[0] or j > r.shape[1]:
        raise IndexError
    output = r[i + offset, j + offset]
    return output


@njit
def P(i, a, b, el, r):
    """P function for rotation matrix calculation.

    Args:
        i (int): index
        a (int): 'a' value
        b (int): 'b' value
        el (int): order
        r (list(matrix)): rotation matrices

    Returns:
        float: P value
    """
    p = 0.0
    if b == el:
        p = (centred_element(r[1], i, 1) * centred_element(r[el - 1], a, el - 1)) - (
            centred_element(r[1], i, -1) * centred_element(r[el - 1], a, -el + 1)
        )
    elif b == -el:
        p = (centred_element(r[1], i, 1) * centred_element(r[el - 1], a, -el + 1)) + (
            centred_element(r[1], i, -1) * centred_element(r[el - 1], a, el - 1)
        )
    else:
        p = centred_element(r[1], i, 0) * centred_element(r[el - 1], a, b)

    return p


@njit
def U(m, n, el, r):
    """U coefficient initialiser for rotation matrix calculation.

    Args:
        m (int): degree
        n (int): index
        el (int): order
        r (list(matrix)): rotation matrices

    Returns:
        float: U value
    """
    return P(0, m, n, el, r)


@njit
def V(m, n, el, r):
    """V coefficient initialiser for rotation matrix calculation.

    Args:
        m (int): degree
        n (int): index
        el (int): order
        r (list(matrix)): rotation matrices

    Returns:
        float: V value
    """
    d = 0
    if m == 0:
        return P(1, 1, n, el, r) + P(-1, -1, n, el, r)
    elif m > 0:
        if m == 1:
            d = 1.0
        return P(1, m - 1, n, el, r) * np.sqrt(1 + d) - P(-1, -m + 1, n, el, r) * (
            1 - d
        )
    else:
        if m == -1:
            d = 1.0
        return P(1, m + 1, n, el, r) * (1 - d) + P(-1, -m - 1, n, el, r) * np.sqrt(
            1 + d
        )


@njit
def W(m, n, el, r):
    """W coefficient initialiser for rotation matrix calculation.

    Args:
        m (int): degree
        n (int): index
        el (int): order
        r (list(matrix)): rotation matrices

    Returns:
        float: W value
    """
    if m == 0:
        return 0.0
    elif m > 0:
        return P(1, m + 1, n, el, r) + P(-1, -m - 1, n, el, r)
    else:
        return P(1, m - 1, n, el, r) - P(-1, -m + 1, n, el, r)


@njit
def compute_UVW_coefficients(m, n, el):
    """Compute U, V and W coefficients for rotation matrix calculation.

    Args:
        m (index): degree
        n (index): index
        el (index): order

    Returns:
        tuple: u, v, w
    """
    d = 0
    if m == 0:
        d = 1
    denom = float((el + n) * (el - n))
    if np.abs(n) == el:
        denom = float(2 * el * (2 * el - 1))

    one_over_denom = 1.0 / denom
    u = np.sqrt(float((el + m) * (el - m)) * one_over_denom)
    v = (
        0.5
        * np.sqrt(
            (1.0 + d)
            * float(el + np.abs(m) - 1)
            * (float(el + np.abs(m)))
            * one_over_denom
        )
        * (1.0 - 2.0 * d)
    )
    w = (
        -0.5
        * np.sqrt(float(el - np.abs(m) - 1) * float(el - np.abs(m)) * one_over_denom)
        * (1.0 - d)
    )

    return u, v, w


@njit
def compute_band_rotation(el, rotations, output):
    """Compute submatrix for rotation matrix.

    Args:
        el (int): order of submatrix
        rotations (list(matrix)): previous and current submatrices
        output (matrix): output destination

    Returns:
        matrix: rotation submatrix
    """
    # print(f'entering band rotation with l = {el}')

    for mm, m in enumerate(np.arange(-el, el + 1, 1)):
        for nn, n in enumerate(np.arange(-el, el + 1, 1)):
            u, v, w = compute_UVW_coefficients(m, n, el)
            if np.abs(u) > 0.0:
                uu = U(m, n, el, rotations)
                u *= uu
            if np.abs(v) > 0.0:
                vv = V(m, n, el, rotations)
                v *= vv
            if np.abs(w) > 0.0:
                ww = W(m, n, el, rotations)
                w *= ww
            rotations[el][mm, nn] = u + v + w

    starting_index = el * el

    output[
        starting_index : (starting_index + (el * 2 + 1)),
        starting_index : (starting_index + (el * 2 + 1)),
    ] = rotations[el]

    return output, rotations


@njit
def dot(A, B):
    """Wraps np.dot for numba #@njit.

    Args:
        A (Array)
        B (Array)

    Returns:
        Array: output
    """
    return np.dot(A, B)


class HOARotator:
    """Provides methods for rotating ambisonics."""

    def __init__(self, order, resolution):
        """Initialize HOARotator."""
        self.order = order
        self.resolution = resolution

        thetas = np.arange(0, 360, resolution)
        n_channels = (order + 1) ** 2

        self.rotmat = np.empty((len(thetas), n_channels, n_channels))

        # NOTE: We are rotating about "z" but construct the rotation matrix
        # with from_euler("y",...) because compute_rotation_matrix
        # is based on Google Resonance Audio code that is y-up.
        #
        # NOTE: the inv is only there to reverse the direction and could
        # be done more easily using -x_value.  The reverse is needed because
        # turning the head 10 degrees to the right is equiv to rotating the room
        # 10 degrees to the left. i.e. rotations given as head movements but
        # are being effected as room movements.
        for i, theta in enumerate(thetas):
            foa_rotmat = R.from_euler("y", theta, degrees=True).as_matrix()
            foa_rotmat = np.linalg.inv(foa_rotmat)
            self.rotmat[i, :, :] = compute_rotation_matrix(order, foa_rotmat)

    def rotate(self, signal, th):
        """Apply rotation to HOA signals using precomputed rotation matrices.

        Args:
            signal (array-like): ambisonic signals
            th (array-like): rotation vector (in radians)

        Returns:
            array-like: transformed ambisonic signals
        """
        # Convert to lookup table indices
        # NOTE: Using generators below - using list comprehensions is *much* slower
        theta_i = (th / np.pi * 180) / self.resolution  # convert to index
        theta_0 = np.floor(theta_i).astype(int)
        n_entries = self.rotmat.shape[0]  # size of table
        t_r0 = (self.rotmat[t % n_entries] for t in theta_0)
        t_r1 = (self.rotmat[(t + 1) % n_entries] for t in theta_0)
        # Interpolate between the two nearest rotation matrices
        t_interp = (
            t_r0_ + alpha_ * (t_r1_ - t_r0_)
            for t_r0_, t_r1_, alpha_ in zip(t_r0, t_r1, theta_i - theta_0)
        )
        # Apply the interpolated rotation matrices to the signal
        signal = np.array(
            [np.dot(x, _t_interp) for x, _t_interp in zip(signal, t_interp)]
        )
        return signal


def binaural_mixdown(ambisonic_signals, hrir, hrir_metadata):
    """Perform binaural mixdown of ambisonic signals.

    Args:
        ambisonic_signals (array-like): inputs
        hrir_filename (string): name of HRIR file
        hrir_metadata (dict): data for channel selection and ambisonic decoding

    Returns:
        array: stereo audio
    """
    # weights = np.array(hrir_metadata["weights"])
    matrix = np.array(hrir_metadata["matrix"])
    logger.info("decoding signal with shape %s", ambisonic_signals.shape)
    logger.info("Decoding to %s positions", matrix.shape[0])

    # Decode to loudspeaker positions

    # # No need to apply max-rE weights given we are using virtual speakers and headphones
    # y = np.dot(ambisonic_signals * weights, inv_matrix)

    n_chans = ambisonic_signals.shape[1]
    inv_matrix = np.linalg.pinv(matrix[:, 0:n_chans])
    y = np.dot(ambisonic_signals, inv_matrix)

    z = np.zeros([y.shape[0] + hrir["M_data"].shape[0] - 1, 2])
    hrir_data = hrir["M_data"][:, hrir_metadata["selected_channels"], :]
    for i in np.arange(y.shape[1]):
        z[:, 0] += convolve(y[:, i], hrir_data[:, i, 0])
        z[:, 1] += convolve(y[:, i], hrir_data[:, i, 1])

    return z


def ambisonic_convolve(signal: np.ndarray, ir: np.ndarray, order: int) -> np.ndarray:
    """Convolve HOAIRs with signals.

    Args:
        signal (ndarray[samples]): the signal to convole
        ir (ndarray[samples, channels]): the HOA impulse responses
        order (int, optional): ambisonic order.

    Returns:
        np.ndarray[samples, channels]: the convolved signal
    """
    n = (order + 1) ** 2
    return np.array([convolve(ir_, signal) for ir_ in ir[:, 0:n].T]).T


def compute_rms(input_signal: np.ndarray, axis: int = 0):
    """Compute rms values along a given axis.
    Args:
        input_signal (np.ndarray): Input signal
        axis (int): Axis along which to compute the Root Mean Square. 0 (default) or 1.

    Returns:
        float: Root Mean Square for the given axis."""

    return np.sqrt(np.mean(input_signal**2, axis=axis))


def equalise_rms_levels(inputs: List[np.ndarray]) -> List[np.ndarray]:
    """Equalise RMS levels.

    Args:
        inputs (array): signals

    Returns:
        array: normalised signals
    """
    rms = compute_rms(np.array(inputs)[:, :, 0], axis=1)
    levels = rms / np.max(rms)
    outputs = [input / level for level, input in zip(levels, inputs)]
    return outputs


def dB_to_gain(x: float) -> float:
    """Convert dB to gain.

    Args:
        x (float):

    Returns
        float:
    """
    return 10 ** (0.05 * x)


def smoothstep(
    x: np.ndarray, x_min: float = 0, x_max: float = 1, N: int = 1
) -> np.ndarray:
    """Apply the smoothstep function.

    Args:
        x (np.ndarray): input
        x_min (float, optional): clamp minimum. Defaults to 0.
        x_max (float, optional): clamp maximum. Defaults to 1.
        N (int, optional): smoothing factor. Defaults to 1.

    Returns:
        np.ndarray: smoothstep values
    """
    x = np.clip((x - x_min) / (x_max - x_min), 0, 1)

    result = sum(
        comb(N + n, n) * comb(2 * N + 1, N - n) * (-x) ** n for n in range(0, N + 1)
    )

    result *= x ** (N + 1)

    return result


def rotation_control_vector(
    array_length: int, start_idx: int, end_idx: int, smoothness: int = 1
) -> np.ndarray:
    """Generate mapped rotation control vector for values of theta.

    Args:
        array_length (int)
        start_idx (int)
        end_idx (int)
        smoothness (int, optional) Defaults to 1.

    Returns:
        array: mapped rotation control vector
    """
    # assert end_idx > start_idx
    # assert array_length > end_idx
    if start_idx > end_idx:
        raise ValueError(f"start_idx ({start_idx}) > end_idx ({end_idx})")
    if end_idx > array_length:
        raise ValueError(f"array_length ({array_length}) > end_idx {end_idx}")
    x = np.arange(0, array_length)
    idx = smoothstep(x, x_min=start_idx, x_max=end_idx, N=smoothness)
    idx = np.array(np.floor(idx * (array_length - 1)), dtype=int)
    return idx


def rotation_vector(
    start_angle: float,
    end_angle: float,
    signal_length: int,
    start_idx: int,
    end_idx: int,
) -> np.ndarray:
    """Compute the rotation vector.

    Args:
        start_angle (float)
        end_angle (float)
        signal_length (int)
        start_idx (int)
        end_idx (int)

    Returns:
        np.array: _description_
    """
    turn_direction = -np.sign(start_angle - end_angle)
    with np.errstate(divide="raise"):
        try:
            increment = (
                np.abs(start_angle - end_angle) / signal_length
            ) * turn_direction
        except FloatingPointError:
            raise
    theta = np.arange(start_angle, end_angle, increment)
    idx = rotation_control_vector(signal_length, start_idx, end_idx)
    return theta[idx]



import copy
import logging

import numpy as np
from scipy.interpolate import interp1d


def compute_proportion_overlap(a1, a2, b1, b2):
    """Compute proportion of overlap of two ranges.

    For ranges (a1, a2) and (b1, b2), express the extent of the overlap
    in the range as a proportion of the extent of range (b1, b2).add()

    e.g (4, 9) and (6, 15) -> overlap (6,9), so proportion is (9-6)/(15-6)

    """
    left = max(a1, b1)
    right = min(a2, b2)
    if left > right:
        overlap = 0.0
    else:
        overlap = float(right - left) / (b2 - b1)
    return overlap


def isothr(vsDesF):
    """Calculate conversion factors of HL thresholds to SPL thresholds.

    Translation of OpenMHA isothr.m. Calculates conversion factors of HL
    thresholds to SPL thresholds. Values from 20 Hz to 12500 Hz are taken
    from ISO 226:2003(E). Values from 14000 Hz to 18000 Hz are taken from ISO
    389-7:2005 (reference thresholds of hearing for free field listening).
    Values at 0 and 20000 Hz are not taken from the ISO Threshold contour.

    Args:
        vsDesF (list): centre frequencies for the amplification bands as 177,
            297, 500, 841,  1414,  2378, 4000, 6727, 11314 Hz

    Returns:
        ndarray: conversion factors

    """
    iso226_389 = np.zeros((34, 2))
    iso226_389[:, 0] = [
        0,
        20,
        25,
        31.5,
        40,
        50,
        63,
        80,
        100,
        125,
        160,
        200,
        250,
        315,
        400,
        500,
        630,
        800,
        1000,
        1250,
        1600,
        2000,
        2500,
        3150,
        4000,
        5000,
        6300,
        8000,
        10000,
        12500,
        14000,
        16000,
        18000,
        20000,
    ]
    iso226_389[:, 1] = [
        80.0,
        78.5,
        68.7,
        59.5,
        51.1,
        44.0,
        37.5,
        31.5,
        26.5,
        22.1,
        17.9,
        14.4,
        11.4,
        8.6,
        6.2,
        4.4,
        3.0,
        2.2,
        2.4,
        3.5,
        1.7,
        -1.3,
        -4.2,
        -6.0,
        -5.4,
        -1.5,
        6.0,
        12.6,
        13.9,
        12.3,
        18.4,
        40.2,
        73.2,
        70.0,
    ]
    vThr = iso226_389[:, 1]
    vsF = iso226_389[:, 0]

    if isinstance(vsDesF, list):
        vsDesF = np.array(vsDesF)

    if np.size(vsDesF[vsDesF < 50]) > 0:
        logging.warning("Frequency values below 50 Hz set to 50 Hz")
        vsDesF[vsDesF < 50] = 50

    vIsoThrDB = interp1d(vsF, vThr, fill_value="extrapolate")(vsDesF)

    return vIsoThrDB


def freq_interp_sh(f_in, y_in, f):
    """Linear interpolation on logarithmic frequency scale.

    Has samples and hold on edges.

    Args:
        f_in (ndarray): audiogram frequencies (Hz)
        y_in (ndarray): audiogram levels
        f (list): FFT filterbank frequencies

    Returns:
        ndarray: interpolated levels corresponding to filterbank frequencies

    """

    # Checks
    if np.size(f[0]) > 1:
        f = f[0]
    if np.size(f_in[0]) > 1:
        f_in = f_in[0]
    if np.size(y_in[0]) > 1:
        y_in = y_in[0]

    vals = np.pad(
        f_in.astype(float), 1, constant_values=((0.5 * f_in[0], 2 * f_in[-1]))
    )
    yvals = np.pad(y_in, 1, constant_values=((y_in[0], y_in[-1])))

    y = interp1d(np.log(vals), yvals, fill_value="extrapolate")(np.log(f))
    y = np.expand_dims(y, 0)

    return y


def gains(compr_thr_inputs, compr_thr_gains, compression_ratios, levels):
    """Based on OpenMHA gains subfunction of gainrule_camfit_compr.

    Args:
        compr_thr_inputs (ndarray): levels for speech in dynamic compression (dc)
            bands minus minima distance (38 dB)
        compr_thr_gains (ndarray): interpolated audiogram levels plus conversion
            factors of HL thresholds to SPL thresholds (output of isothr) minus
            compr_thr_inputs
        compression_ratios (ndarray): compression ratios according to CAMFIT compressive
        levels (ndarray): set of levels over which to calculate gains e.g. -10:110 dB

    Returns:
        ndarray: set of uncorrected gains as 2-d numpy array

    """
    levels = np.transpose(np.tile(levels, (np.size(compr_thr_inputs), 1)))
    compr_thr_inputs = np.tile(compr_thr_inputs, (np.shape(levels)[0], 1))
    compr_thr_gains = np.tile(compr_thr_gains, (np.shape(levels)[0], 1))
    compression_ratios = np.tile(compression_ratios, (np.shape(levels)[0], 1))

    compr_thr_outputs = compr_thr_inputs + compr_thr_gains
    outputs = (levels - compr_thr_inputs) / compression_ratios + compr_thr_outputs
    g = outputs - levels

    return g


def gainrule_camfit_linear(
    audiogram, sFitmodel, noisegatelevels=45, noisegateslope=1, max_output_level=100
):
    """Apply linear Cambridge rule for hearing aid fittings 'CAMFIT'.

    Based on OpenMHA gainrule_camfit_linear.m. Applies linear Cambridge rule for
    hearing aid fittings 'CAMFIT'. Implemented as described in B. Moore and B.
    Glasberg (1998), "Use of a loudness model for hearing-aid fitting. I. Linear
    hearing aids" Brit. J. Audiol. (32) 317-335.

    The gain rule limits the gains so that in each band 100 dB output level
    is not exceeded.
    The Cambridge formula defines intercepts only up to 5 kHz. Because the
    intercepts do not vary much between 1kHz and 5kHz anyway (these are all
    within 0dB +/- 1dB), we extend the last intercept of +1dB at 5kHz to all higher
    frequencies. This function assumes audiogram frequencies are identical for
    two ears.

    The original function is part of the HörTech Open Master Hearing Aid (openMHA)
    Copyright © 2007 2009 2011 2013 2015 2017 2018 HörTech gGmbH
    openMHA is free software: see licencing conditions at http://www.openmha.org/

    Args:
        sAud (dict): contains the subject-specific hearing threshold levels in
            dB HL for the left and right ears, and the audiogram frequencies
        sFitmodel (dict): contains the center frequencies for the amplification
            bands and input levels in SPL for which to compute the gains
        noisegatelevels (ndarray): compression threshold levels for each frequency
            band (default: 45)
        noisegateslope (int): determines slope of gains below compression threshold
            (default: 1)

    Returns:
        dict: dictionary containing gain table, noise gate, and noise
            gate expansion slope fields

    """

    intercept_frequencies = np.array(
        [125, 250, 500, 750, 1000, 1500, 2000, 3000, 4000, 5000, 5005]
    )
    intercepts = np.array([-11, -10, -8, -6, 0, -1, 1, -1, 0, 1, 1])
    try:
        if np.size(sFitmodel["frequencies"][0][0]) > 1:
            intercepts = freq_interp_sh(
                intercept_frequencies, intercepts, sFitmodel["frequencies"][0][0]
            )
    except Exception:  # noqa E722
        intercepts = freq_interp_sh(
            intercept_frequencies, intercepts, sFitmodel["frequencies"]
        )

    sFitmodel_frequencies = sFitmodel["frequencies"]
    sFitmodel_levels = sFitmodel["levels"]
    if np.size(sFitmodel_frequencies) == 1:
        sFitmodel_frequencies = sFitmodel_frequencies[0][0]
        sFitmodel_levels = sFitmodel_levels[0][0]

    if np.shape(sFitmodel_levels)[0] != 1:
        sFitmodel_levels = np.transpose(sFitmodel_levels)

    # Interpolate audiogram
    # num levels x num freqs x L, R
    sGt = np.zeros((np.size(sFitmodel_levels), np.size(sFitmodel_frequencies), 2))

    noisegate_level = np.zeros((np.size(sFitmodel_frequencies), 2))
    noisegate_slope = np.zeros((np.size(sFitmodel_frequencies), 2))
    insertion_gains_out = np.zeros((np.size(sFitmodel_frequencies), 2))

    for i, levels in enumerate([audiogram.levels_l, audiogram.levels_r]):
        htlside = freq_interp_sh(audiogram.cfs, levels, sFitmodel_frequencies)

        insertion_gains = htlside * 0.48 + intercepts

        # According to B. Moore (1998), "Use of a loudness model for hearing-aid
        # fitting. II. Hearing aids with multi-channel compression" Brit. J.
        # Audiol. (33) 157-170, p. 159, do not permit negative insertion gains in
        # practice.
        insertion_gains[insertion_gains < 0] = 0

        # Set all gains to 0 for 0dB HL flat audiogram
        insertion_gains = insertion_gains * htlside.any()  # any(htlside)

        sGt[:, :, i] = np.tile(insertion_gains, (len(sFitmodel_levels), 1))

        if np.size(insertion_gains[0]) == np.size(htlside):
            insertion_gains = insertion_gains[0]

        insertion_gains_out[:, i] = insertion_gains
        output_levels = np.tile(sFitmodel_levels, (np.size(insertion_gains), 1))
        output_levels = sGt[:, :, i] + np.transpose(output_levels)

        # Where output level is greater than max_output_level, reduce gain
        safe_output_levels = copy.deepcopy(output_levels)
        safe_output_levels[safe_output_levels > max_output_level] = max_output_level

        sGt[:, :, i] = sGt[:, :, i] - (output_levels - safe_output_levels)

        noisegate_level[:, i] = noisegatelevels * np.ones(
            np.size(sFitmodel_frequencies)
        )
        noisegate_slope[:, i] = noisegateslope * np.ones(np.size(sFitmodel_frequencies))

    # overall_level = 10 * np.log(np.sum(10 ** (insertion_gains_out / 10), axis=0))

    output = {}
    output["sGt"] = sGt
    output["noisegate_level"] = noisegate_level
    output["noisegate_slope"] = noisegate_slope
    output["insertion_gains"] = insertion_gains_out

    return output


def gainrule_camfit_compr(
    audiogram,
    sFitmodel,
    noisegatelevels=45,
    noisegateslope=1,
    level=0,
    max_output_level=100,
):
    """Applies compressive Cambridge rule for hearing aid fittings 'CAMFIT'.

    Translation of OpenMHA gainrule_camfit_compr.m.
    Applies compressive Cambridge rule for
    hearing aid fittings 'CAMFIT'. Computes gains for compression according to Moore
    et al. (1999) "Use of a loudness model for hearing aid fitting: II. Hearing aids
    with multi-channel compression." Brit. J. Audiol. (33) 157-170

    The gain rule limits the gains so that in each band 100 dB output level is
    not exceeded. This function assumes audiogram frequencies are identical for the
    two ears. In this implementation, any negative gains are set to 0 dB.

    The original function is part of the HörTech Open Master Hearing Aid (openMHA)
    Copyright © 2007 2009 2011 2013 2015 2017 2018 HörTech gGmbH
    openMHA is free software: see licencing conditions at http://www.openmha.org/

    Args:
        audiogram (Audiogram): the audiogram for which to make the fit
        sFitmodel (dict): contains the center frequencies for the amplification
            bands and input levels in SPL for which to compute the gains
        noisegatelevels (ndarray): compression threshold levels for each frequency
            band (default: 45)
        noisegateslope (int): determines slope of gains below compression threshold
        level (int): input level in each band for compression rate calculation
            (default: 0 for variable level depending on insertion gains)
        max_output_level (int): maximum output level in dB

    Returns:
        dict: dictionary containing gain table, noise gate, and noise
            gate expansion slope fields

    """

    # International long-term average speech spectrum for speech with an overall
    # level of 70dB in third-octave frequency bands, taken from Byrne et al.
    # (1994) J. Acoust. Soc. Am. 96(4) 2108-2120
    # Average across males and females SG
    LTASS_freq = np.array(
        [
            63,
            80,
            100,
            125,
            160,
            200,
            250,
            315,
            400,
            500,
            630,
            800,
            1000,
            1250,
            1600,
            2000,
            2500,
            3150,
            4000,
            5000,
            6300,
            8000,
            10000,
            12500,
            16000,
        ]
    )
    LTASS_edge_freq = np.zeros((26))
    LTASS_edge_freq[1:-1] = np.sqrt(LTASS_freq[0:-1] * LTASS_freq[1:])
    LTASS_edge_freq[-1] = 16000 * np.power(2, (1 / 6))
    LTASS_lev = np.array(
        [
            38.6,
            43.5,
            54.4,
            57.7,
            56.8,
            60.2,
            60.3,
            59.0,
            62.1,
            62.1,
            60.5,
            56.8,
            53.7,
            53.0,
            52.0,
            48.7,
            48.1,
            46.8,
            45.6,
            44.5,
            44.3,
            43.7,
            43.4,
            41.3,
            40.7,
        ]
    )
    LTASS_intensity = np.power(10, LTASS_lev / 10)

    frequencies = np.array(sFitmodel["frequencies"])
    edge_freq = np.array(sFitmodel["edge_frequencies"])

    sFitmodel_frequencies = np.array(sFitmodel["frequencies"])
    sFitmodel_levels = np.array(sFitmodel["levels"])

    speech_level_65_in_dc_bands = np.zeros(np.shape(frequencies))

    for band, (f_range_a, f_range_b) in enumerate(zip(edge_freq[:-1], edge_freq[1:])):
        portion = [
            compute_proportion_overlap(f_range_a, f_range_b, ltass_a, ltass_b)
            for (ltass_a, ltass_b) in zip(LTASS_edge_freq[:-1], LTASS_edge_freq[1:])
        ]
        intensity_sum = np.inner(LTASS_intensity, portion)  # weighted sum
        speech_level_70_in_dc_band = 10 * np.log10(intensity_sum)
        speech_level_65_in_dc_bands[band] = speech_level_70_in_dc_band - 5

    # minima in lowest level speech that needs to be understood is 38 dB below
    # speech_level_65_in_dc_bands
    minima_distance = 38

    # Conversion factors of HL thresholds to SPL thresholds
    Conv = isothr(frequencies)

    # Get speech minima
    Lmin = speech_level_65_in_dc_bands - minima_distance

    # Interpolate audiogram and get absolute thresholds in dB HL at centre frequencies
    # and gains required for speech minima
    htl = np.zeros((np.size(frequencies), 2))
    Gmin = np.zeros(np.shape(htl))

    for i, levels in enumerate([audiogram.levels_l, audiogram.levels_r]):
        htl[:, i] = freq_interp_sh(audiogram.cfs, levels, frequencies)
        Gmin[:, i] = htl[:, i] + Conv - Lmin

    # Get input levels
    Lmid = speech_level_65_in_dc_bands

    # Calculate gains at centre frequencies
    Gmid = gainrule_camfit_linear(
        audiogram, sFitmodel, noisegatelevels, noisegateslope, max_output_level
    )
    insertion_gains = Gmid["insertion_gains"]
    Gmid = Gmid["sGt"]

    # Calculate compression ratios
    compression_ratio = np.zeros((np.size(frequencies), 2))
    sGt = np.zeros((len(sFitmodel_levels), len(sFitmodel_frequencies), 2))

    noisegate_level = np.zeros((np.size(sFitmodel_frequencies), 2))
    noisegate_slope = np.zeros((np.size(sFitmodel_frequencies), 2))

    # Find index corresponding to input level in dB for compression rate calculation
    if level != 0:
        cr_idx = [i for (i, val) in enumerate(sFitmodel_levels) if val == level]

    for i, levels in enumerate([audiogram.levels_l, audiogram.levels_r]):
        if level != 0:
            tmp = Lmid + Gmid[cr_idx, :, i].flatten() - Lmin - Gmin[:, i]
        else:
            tmp = Lmid + insertion_gains[:, i] - Lmin - Gmin[:, i]
        idx = [i for i, x in enumerate(tmp < 13) if x]
        tmp[idx] = 13
        compression_ratio[:, i] = minima_distance / tmp
        compression_ratio[:, i][compression_ratio[:, i] < 1] = 1

        sGt[:, :, i] = gains(
            Lmin, Gmin[:, i], compression_ratio[:, i], sFitmodel_levels
        )
        # Set negative gains to zero
        sGt[:, :, i][sGt[:, :, i] < 0] = 0
        # where output level is greater than max_output_level, reduce gain
        tmp = np.tile(sFitmodel_levels, (len(Gmin[:, i]), 1))
        tmp = np.transpose(tmp)
        output_levels = sGt[:, :, i] + tmp

        safe_output_levels = copy.deepcopy(output_levels)
        safe_output_levels[safe_output_levels > max_output_level] = max_output_level
        sGt[:, :, i] = sGt[:, :, i] - (output_levels - safe_output_levels)

        # set all gains to 0 for 0dB HL flat audiogram
        sGt[:, :, i] = sGt[:, :, i] * levels.any()

        noisegate_level[:, i] = noisegatelevels * np.ones(
            np.size(sFitmodel_frequencies)
        )
        noisegate_slope[:, i] = noisegateslope * np.ones(np.size(sFitmodel_frequencies))

    logging.info("Noisegate levels are %s", noisegate_level)

    output = {}
    output["sGt"] = sGt
    output["noisegatelevel"] = noisegate_level
    output["noisegateslope"] = noisegate_slope

    return output



"""Modified Binaural Short-Time Objective Intelligibility (MBSTOI) Measure"""
import importlib.resources as pkg_resources
import logging
import math

import numpy as np
import yaml  # type: ignore
from scipy.signal import resample

from clarity.evaluator.mbstoi.mbstoi_utils import (
    equalisation_cancellation,
    remove_silent_frames,
    stft,
    thirdoct,
)

# pylint: disable=too-many-locals


# basic stoi parameters from file
params_file = pkg_resources.open_text(__package__, "parameters.yaml")
basic_stoi_parameters = yaml.safe_load(params_file.read())


def mbstoi(
    left_ear_clean: np.ndarray,
    right_ear_clean: np.ndarray,
    left_ear_noisy: np.ndarray,
    right_ear_noisy: np.ndarray,
    fs_signal,
    gridcoarseness: int = 1,
    sample_rate: int = 10000,
    n_frame: int = 256,
    fft_size_in_samples: int = 512,
    n_third_octave_bands: int = 15,
    centre_freq_first_third_octave_hz: int = 150,
    n_frames: int = 30,
    dyn_range: int = 40,
    tau_min: float = -0.001,
    tau_max: float = 0.001,
    gamma_min: int = -20,
    gamma_max: int = 20,
    sigma_delta_0: float = 65e-6,
    sigma_epsilon_0: float = 1.5,
    alpha_0_db: int = 13,
    tau_0: float = 1.6e-3,
    level_shift_deviation: float = 1.6,
) -> float:
    """Implementation of the Modified Binaural Short-Time Objective Intelligibility (mbstoi) measure.

    Args:
        left_ear_clean (np.ndarray): Clean speech signal from left ear.
        right_ear_clean (np.ndarray): Clean speech signal from right ear.
        left_ear_noisy (np.ndarray) : Noisy/processed speech signal from left ear.
        right_ear_noisy (np.ndarray) : Noisy/processed speech signal from right ear.
        fs_signal (int) : Frequency sample rate of signal.
        gridcoarseness (int) : Grid coarseness as denominator of ntaus and ngammas (default: 1).
        sample_rate (int) :  Sample Rate.
        n_frame (int) :  Number of Frames.
        fft_size_in_samples (int) :  ??? size in samples.
        n_third_octave_bands (int) : Number of third octave bands.
        centre_freq_first_third_octave_hz (int) :  150,
        n_frames (int) :  Number of Frames.
        dyn_range (int) : Dynamic Range.
        tau_min (float) : Min Tau the ???
        tau_max (float) : Max Tau the ???
        gamma_min (int) : Minimum gamma the ???
        gamma_max (int) : Maximum gamma the ???
        sigma_delta_0 (float) : ???
        sigma_epsilon_0 (float) : ???
        alpha_0_db (int) : ???
        tau_0 (float) : ???
        level_shift_deviation (float) : ???

    Returns:
        float : mbtsoi index d.

    Notes:
        All title, copyrights and pending patents pertaining to mbtsoi[1]_ in and to the original Matlab software are
        owned by oticon a/s and/or Aalborg University. please see details at
        `http://ah-andersen.net/code/<http://ah-andersen.net/code/>`


    .. [1] A. H. Andersen, J. M. de Haan, Z.-H. Tan, and J. Jensen (2018) Refinement and validation of the binaural
    short time objective intelligibility measure for spatially diverse conditions. Speech Communication vol. 102,
    pp. 1-13 `doi:10.1016/j.specom.2018.06.001 <https://doi.org/10.1016/j.specom.2018.06.001>`_
    """

    n_taus = math.ceil(100 / gridcoarseness)  # number of tau values to try out
    n_gammas = math.ceil(40 / gridcoarseness)  # number of gamma values to try out

    # prepare signals, ensuring that inputs are column vectors
    left_ear_clean = left_ear_clean.flatten()
    right_ear_clean = right_ear_clean.flatten()
    left_ear_noisy = left_ear_noisy.flatten()
    right_ear_noisy = right_ear_noisy.flatten()

    # Resample signals to 10 kHz
    if fs_signal != sample_rate:

        logging.debug(
            "Resampling signals with sr=%s for MBSTOI calculation.", sample_rate
        )
        # Assumes fs_signal is 44.1 kHz
        length_left_ear_clean = len(left_ear_clean)
        left_ear_clean = resample(
            left_ear_clean, int(length_left_ear_clean * (sample_rate / fs_signal) + 1)
        )
        right_ear_clean = resample(
            right_ear_clean, int(length_left_ear_clean * (sample_rate / fs_signal) + 1)
        )
        left_ear_noisy = resample(
            left_ear_noisy, int(length_left_ear_clean * (sample_rate / fs_signal) + 1)
        )
        right_ear_noisy = resample(
            right_ear_noisy, int(length_left_ear_clean * (sample_rate / fs_signal) + 1)
        )

    # Remove silent frames
    (
        left_ear_clean,
        right_ear_clean,
        left_ear_noisy,
        right_ear_noisy,
    ) = remove_silent_frames(
        left_ear_clean,
        right_ear_clean,
        left_ear_noisy,
        right_ear_noisy,
        dyn_range,
        n_frame,
        n_frame / 2,
    )

    # Handle case when signals are zeros
    if (
        abs(np.log10(np.linalg.norm(left_ear_clean) / np.linalg.norm(left_ear_noisy)))
        > 5.0
        or abs(
            np.log10(np.linalg.norm(right_ear_clean) / np.linalg.norm(right_ear_noisy))
        )
        > 5.0
    ):
        sii = 0

    # STDFT and filtering
    # Get 1/3 octave band matrix
    [
        octave_band_matrix,
        centre_frequencies,
        frequency_band_edges_indices,
        _freq_low,
        _freq_high,
    ] = thirdoct(
        sample_rate,
        fft_size_in_samples,
        n_third_octave_bands,
        centre_freq_first_third_octave_hz,
    )  # (fs, nfft, num_bands, min_freq)
    centre_frequencies = (
        2 * math.pi * centre_frequencies
    )  # This is now the angular frequency in radians per sec

    # Apply short time DFT to signals and transpose
    left_ear_clean_hat = stft(left_ear_clean, n_frame, fft_size_in_samples).transpose()
    right_ear_clean_hat = stft(
        right_ear_clean, n_frame, fft_size_in_samples
    ).transpose()
    left_ear_noisy_hat = stft(left_ear_noisy, n_frame, fft_size_in_samples).transpose()
    right_ear_noisy_hat = stft(
        right_ear_noisy, n_frame, fft_size_in_samples
    ).transpose()

    # Take single sided spectrum of signals
    idx = int(fft_size_in_samples / 2 + 1)
    left_ear_clean_hat = left_ear_clean_hat[0:idx, :]
    right_ear_clean_hat = right_ear_clean_hat[0:idx, :]
    left_ear_noisy_hat = left_ear_noisy_hat[0:idx, :]
    right_ear_noisy_hat = right_ear_noisy_hat[0:idx, :]

    # Compute intermediate correlation via EC search
    logging.info("Starting EC evaluation")
    # Here intermeduiate correlation coefficients are evaluated for a discrete set of
    # gamma and tau values (a "grid") and the highest value is chosen.
    intermediate_intelligibility_measure_grid = np.zeros(
        (n_third_octave_bands, np.shape(left_ear_clean_hat)[1] - n_frames + 1)
    )
    p_ec_max = np.zeros(
        (n_third_octave_bands, np.shape(left_ear_clean_hat)[1] - n_frames + 1)
    )

    # Interaural compensation time and level values
    taus = np.linspace(tau_min, tau_max, n_taus)
    gammas = np.linspace(gamma_min, gamma_max, n_gammas)

    # Jitter incorporated below - Equations 5 and 6 in Andersen et al. 2018
    sigma_epsilon = (
        np.sqrt(2)
        * sigma_epsilon_0
        * (1 + (abs(gammas) / alpha_0_db) ** level_shift_deviation)
        / 20
    )
    gammas = gammas / 20

    sigma_delta = np.sqrt(2) * sigma_delta_0 * (1 + (abs(taus) / tau_0))

    logging.info("Processing Equalisation Cancellation stage")
    updated_intermediate_intelligibility_measure, p_ec_max = equalisation_cancellation(
        left_ear_clean_hat,
        right_ear_clean_hat,
        left_ear_noisy_hat,
        right_ear_noisy_hat,
        n_third_octave_bands,
        n_frames,
        frequency_band_edges_indices,
        centre_frequencies.flatten(),
        taus,
        n_taus,
        gammas,
        n_gammas,
        intermediate_intelligibility_measure_grid,
        p_ec_max,
        sigma_epsilon,
        sigma_delta,
    )

    # Compute the better ear STOI
    logging.info("Computing better ear intermediate correlation coefficients")
    # Arrays for the 1/3 octave envelope
    left_ear_clean_third_octave_band = np.zeros(
        (n_third_octave_bands, np.shape(left_ear_clean_hat)[1])
    )
    right_ear_clean_third_octave_band = np.zeros(
        (n_third_octave_bands, np.shape(left_ear_clean_hat)[1])
    )
    left_ear_noisy_third_octave_band = np.zeros(
        (n_third_octave_bands, np.shape(left_ear_clean_hat)[1])
    )
    right_ear_noisy_third_octave_band = np.zeros(
        (n_third_octave_bands, np.shape(left_ear_clean_hat)[1])
    )

    # Apply 1/3 octave bands as described in Eq.(1) of the STOI article
    for k in range(np.shape(left_ear_clean_hat)[1]):
        left_ear_clean_third_octave_band[:, k] = np.dot(
            octave_band_matrix, abs(left_ear_clean_hat[:, k]) ** 2
        )
        right_ear_clean_third_octave_band[:, k] = np.dot(
            octave_band_matrix, abs(right_ear_clean_hat[:, k]) ** 2
        )
        left_ear_noisy_third_octave_band[:, k] = np.dot(
            octave_band_matrix, abs(left_ear_noisy_hat[:, k]) ** 2
        )
        right_ear_noisy_third_octave_band[:, k] = np.dot(
            octave_band_matrix, abs(right_ear_noisy_hat[:, k]) ** 2
        )

    # Arrays for better-ear correlations
    dl_interm = np.zeros(
        (n_third_octave_bands, len(range(n_frames, len(left_ear_clean_hat[1]) + 1)))
    )
    dr_interm = np.zeros(
        (n_third_octave_bands, len(range(n_frames, len(left_ear_clean_hat[1]) + 1)))
    )
    left_improved = np.zeros(
        (n_third_octave_bands, len(range(n_frames, len(left_ear_clean_hat[1]) + 1)))
    )
    right_improved = np.zeros(
        (n_third_octave_bands, len(range(n_frames, len(left_ear_clean_hat[1]) + 1)))
    )

    # Compute temporary better-ear correlations
    for m in range(
        n_frames, np.shape(left_ear_clean_hat)[1]
    ):  # pylint: disable=invalid-name
        left_ear_clean_seg = left_ear_clean_third_octave_band[:, (m - n_frames) : m]
        right_ear_clean_seg = right_ear_clean_third_octave_band[:, (m - n_frames) : m]
        left_ear_noisy_seg = left_ear_noisy_third_octave_band[:, (m - n_frames) : m]
        right_ear_noisy_seg = right_ear_noisy_third_octave_band[:, (m - n_frames) : m]

        for n in range(n_third_octave_bands):  # pylint: disable=invalid-name
            left_ear_clean_n = (
                left_ear_clean_seg[n, :] - np.sum(left_ear_clean_seg[n, :]) / n_frames
            )
            right_ear_clean_n = (
                right_ear_clean_seg[n, :] - np.sum(right_ear_clean_seg[n, :]) / n_frames
            )
            left_ear_noisy_n = (
                left_ear_noisy_seg[n, :] - np.sum(left_ear_noisy_seg[n, :]) / n_frames
            )
            right_ear_noisy_n = (
                right_ear_noisy_seg[n, :] - np.sum(right_ear_noisy_seg[n, :]) / n_frames
            )
            left_improved[n, m - n_frames] = np.sum(
                left_ear_clean_n * left_ear_clean_n
            ) / np.sum(left_ear_noisy_n * left_ear_noisy_n)
            right_improved[n, m - n_frames] = np.sum(
                right_ear_clean_n * right_ear_clean_n
            ) / np.sum(right_ear_noisy_n * right_ear_noisy_n)
            dl_interm[n, m - n_frames] = np.sum(left_ear_clean_n * left_ear_noisy_n) / (
                np.linalg.norm(left_ear_clean_n) * np.linalg.norm(left_ear_noisy_n)
            )
            dr_interm[n, m - n_frames] = np.sum(
                right_ear_clean_n * right_ear_noisy_n
            ) / (np.linalg.norm(right_ear_clean_n) * np.linalg.norm(right_ear_noisy_n))

    # Get the better ear intermediate coefficients
    idx = np.isfinite(dl_interm)
    dl_interm[~idx] = 0
    idx = np.isfinite(dr_interm)
    dr_interm[~idx] = 0
    p_be_max = np.maximum(left_improved, right_improved)
    dbe_interm = np.zeros((np.shape(dl_interm)))

    idx = left_improved > right_improved
    dbe_interm[idx] = dl_interm[idx]
    dbe_interm[~idx] = dr_interm[~idx]

    # Compute STOI measure
    # Whenever a single ear provides a higher correlation than the corresponding EC
    # processed alternative,the better-ear correlation is used.
    idx = p_be_max > p_ec_max
    updated_intermediate_intelligibility_measure[idx] = dbe_interm[idx]
    sii = np.mean(updated_intermediate_intelligibility_measure)

    logging.info("MBSTOI processing complete")

    return sii



"""Code for building the scenes.json files."""
import itertools
import json
import logging
import math
import random
import re
from enum import Enum

import numpy as np
from tqdm import tqdm

# A logger for this file
log = logging.getLogger(__name__)


# Get json output to round to 4 dp
json.encoder.c_make_encoder = None  # type: ignore


class RoundingFloat(float):
    """Round a float to 4 decimal places."""

    __repr__ = staticmethod(lambda x: format(x, ".4f"))  # type: ignore


json.encoder.float = RoundingFloat  # type: ignore

# rpf file Handling

N_SCENES = 10000  # Number of scenes to expect
N_INTERFERERS = 3  # Default number of interferers to expect


def set_random_seed(random_seed):
    if random_seed is not None:
        random.seed(random_seed)
        np.random.seed(random_seed)


def get_vector(text, vector_name):
    """Get a vector quantity from the rpf file.
    Will read rpf vector quantities, eg.
    "sourceViewVectors = -0.095,-0.995, 0.000"

    Args:
        text (str): string contents of the rpf file
        vector_name (str): name of vector to extract (e.g. "sourceViewVectors")

    Returns:
        list[float]: vector as list of floats
    """
    findall_str = f".*{vector_name}.*"
    x = re.findall(findall_str, text)
    x = re.sub(" ", "", x[0].split("=")[-1])
    x = [float(idx) for idx in x.split(",")]
    return x


def get_room_dims(text: str) -> list:
    """Find the room dimensions in the rpf file.

    Args:
        text (str): String to be searched for room dimensions (string to be searched for is of the form 'ProjectName =
    CuboidRoom_5.9x3.4186x2.9').


    Returns:
        list: List of the three dimensions of the room.
    """
    # Room dimensions appear in the file like this,
    #    ProjectName = CuboidRoom_5.9x3.4186x2.9
    roomdims = re.findall(r"ProjectName = .*", text)[0]
    return roomdims.split("=")[-1].split("_")[-1]


def get_room_name(text: str) -> str:
    """Find the room name in the rpf file.

    Args:
        text (str): String to be searched for room name ('R' followed by 5 digits).

    Returns:
        str: The room name.
    """
    return re.findall(r"R\d\d\d\d\d", text)[0]


def read_rpf_file(rpf_filename):
    """Process an rpf file and return key contents as a dictionary."""
    with open(rpf_filename, "r", encoding="utf-8") as f:
        text = f.read()

    rpf_dict = {}

    room_dict = {}
    room_dict["name"] = get_room_name(text)
    room_dict["dimensions"] = get_room_dims(text)

    rpf_dict["room"] = room_dict

    # This hrir is not needed - added later
    # rpf_dict["hrirfilename"] = get_hrir_filename(text)

    rpf_dict["source"] = {
        "position": get_vector(text, "sourcePositions"),
        "view_vector": get_vector(text, "sourceViewVectors"),
    }

    rpf_dict["receiver"] = {
        "position": get_vector(text, "receiverPositions"),
        "view_vector": get_vector(text, "receiverViewVectors"),
    }

    return rpf_dict


def build_room(target_file, interferer_files):
    """Build room json file from contents of related rpf files.
    Note, there is an rpf file for each source in the scene. All of these
    files are read and a single scene json file is constructed.

    Args:
        target_file (str): rpf file containing the target position
        interferer_files (list[str]): list of files containing the interferer positions

    Returns:
        dict: dictionary representation of the scene following CEC2 scene.json format
    """
    scene_dict = {}
    scene_dict["target"] = {}
    scene_dict["listener"] = {}
    scene_dict["interferer"] = {}

    rpf_dict = read_rpf_file(target_file)

    scene_dict = {}
    scene_dict["name"] = rpf_dict["room"]["name"]
    scene_dict["dimensions"] = rpf_dict["room"]["dimensions"]
    scene_dict["target"] = rpf_dict["source"]
    scene_dict["listener"] = rpf_dict["receiver"]

    # Read interferer rpf files and extract source info
    interferer_rpfs = [
        read_rpf_file(interferer_file) for interferer_file in interferer_files
    ]
    interferers = [interferer_rpf["source"] for interferer_rpf in interferer_rpfs]
    # The source view vectors are removed as interferer are all omnidirectional
    for interferer in interferers:
        del interferer["view_vector"]
    scene_dict["interferers"] = interferers

    return scene_dict


def make_rpf_filename_dict(rpf_location, scene_str, n_interferers):
    """Construct dictionary storing all rpf files that will be processed."""
    rpf_filename_format = "{rpf_location}/{scene_str}_{source}.rpf"

    rpf_filedict = {}
    rpf_filedict["target"] = rpf_filename_format.format(
        rpf_location=rpf_location, scene_str=scene_str, source="t"
    )

    rpf_filedict["interferers"] = [
        rpf_filename_format.format(
            rpf_location=rpf_location, scene_str=scene_str, source=f"i{i}"
        )
        for i in range(1, n_interferers + 1)
    ]
    return rpf_filedict


# target handling
def get_num_pre_samples(pre_samples_range):
    """Number of samples prior to target onset."""
    return random.randint(*pre_samples_range)


def get_num_post_samples(post_samples_range):
    """Number of samples to continue player after target offsets."""
    return random.randint(*post_samples_range)


def add_this_target_to_scene(target, scene, pre_samples_range, post_samples_range):
    """Add the target details to the scene dict.
    Adds given target to given scene. Target details will be taken
    from the target dict but the start time will be
    according to the CEC2 target start time specification.

    Args:
        target (dict): target dict read from target metadata file
        scene (dict): complete scene dict
        pre_samples_range (list): parameters for number of samples prior to target onset
        post_samples_range (list): parameters for number of samples to continue
            player after target offsets
    """
    scene_target = {}
    num_pre_samples = get_num_pre_samples(pre_samples_range)
    num_post_samples = get_num_post_samples(post_samples_range)
    scene_target["name"] = target["wavfile"]
    scene_target["time_start"] = num_pre_samples
    scene_target["time_end"] = num_pre_samples + target["nsamples"]
    scene["target"] = scene_target
    scene["duration"] = num_pre_samples + target["nsamples"] + num_post_samples


# SNR handling
def generate_snr(snr_range):
    """Generate a random SNR."""
    return random.uniform(*snr_range)


# Interferer handling
class InterfererType(Enum):
    """Enum for interferer types."""

    SPEECH = "speech"
    NOISE = "noise"
    MUSIC = "music"


def select_interferer_types(allowed_n_interferers):
    """Select the interferer types to use.
    The number of interferer is drawn randomly from list of allowed valued.
    The type of each is chosen randomly but there is not allowed to be
    more than 1 music source.

    Args:
        allowed_n_interferers (list): list of allowed number of interferers

    Returns:
        list(InterfererType): list of interferer types to use
    """

    def valid(selection):
        """Check if the selection is valid."""
        return selection.count(InterfererType.MUSIC) <= 1

    n_interferers = random.choice(allowed_n_interferers)
    selection = None
    while selection is None or not valid(selection):
        selection = random.choices(list(InterfererType), k=n_interferers)
    return selection


def select_random_interferer(interferers, dataset, required_samples):
    """Randomly select an interferer.
    Interferers stored as list of list. First randomly select a sublist
    then randomly select an item from sublist matching constraints.

    Args:
        interferers (list(list)): interferers as list of lists
        dataset (str): desired data [train, dev, eval]
        required_samples (int): required number of samples

    Raises:
        ValueError: if no suitable interferer is found

    Returns:
        dict: the interferer dict
    """
    interferer_group = random.choice(interferers)
    filtered_interferer_group = [
        i
        for i in interferer_group
        if i["dataset"] == dataset and i["nsamples"] >= required_samples
    ]
    try:
        interferer = random.choice(filtered_interferer_group)
    except IndexError as e:
        raise ValueError(
            f"No suitable interferer found for dataset {dataset} and required samples"
            f" {required_samples}"
        ) from e
    return interferer


def get_random_interferer_offset(interferer, required_samples):
    """Generate a random offset sample for interferer.
    The offset sample is the point within the masker signal at which the interferer
    segment will be extracted. Randomly selected but with care for it not to start
    too late, i.e. such that the required samples would overrun the end of the masker
    signal will be used is taken.

    Args:
        interferer (dict): the interferer metadata
        required_samples (int): number of samples that is going to be required

    Returns:
        int: a valid randomly selected offset
    """
    masker_nsamples = interferer["nsamples"]
    latest_start = masker_nsamples - required_samples
    if latest_start < 0:
        log.error("Interferer %s does not has enough samples.", interferer["ID"])

    assert (
        latest_start >= 0
    )  # This should never happen - mean masker was too short for the scene
    return random.randint(0, latest_start)


def add_interferer_to_scene_inner(
    scene, interferers, number, start_time_range, end_early_time_range
):
    """Randomly select interferers and add them to the given scene.
    A random number of interferers is chosen, then each is given a random type
    selected from the possible speech, nonspeech, music types.
    Interferers are then chosen from the available lists according to the type
    and also taking care to match the scenes 'dataset' field, ie. train, dev, test.
    The interferer data is supplied as a dictionary of lists of lists. The key
    being "speech", "nonspeech", or "music", and the list of list being a partitioned
    list of interferers for that type.
    The idea of using a list of lists is that interferers can be split by
    subcondition and then the randomization draws equally from each subcondition,
    e.g. for nonspeech there is "washing machine", "microwave" etc. This ensures that
    each subcondition is equally represented even if the number of exemplars of
    each subcondition is different.
    Note, there is no return. The scene is modified in place.

    Args:
        scene (dict): the scene description
        interferers (dict): the interferer metadata
        number: number of interferers
        start_time_range: when to start
        end_early_time_range: when to end
    """
    dataset = scene["dataset"]
    selected_interferer_types = select_interferer_types(number)
    n_interferers = len(selected_interferer_types)
    n_positions_available = len(scene["room"]["interferers"])
    available_positions = range(1, n_positions_available + 1)
    positions = random.sample(available_positions, n_interferers)
    positions.sort()

    # Make list of empty dicts match the number selected
    scene["interferers"] = [{"position": position} for position in positions]

    # Randomly instantiate each interferer in the scene
    for scene_interferer, scene_type in zip(
        scene["interferers"], selected_interferer_types
    ):
        desired_start_time = random.randint(*start_time_range)

        scene_interferer["time_start"] = min(scene["duration"], desired_start_time)
        desired_end_time = scene["duration"] - random.randint(*end_early_time_range)

        scene_interferer["time_end"] = max(
            scene_interferer["time_start"], desired_end_time
        )

        required_samples = scene_interferer["time_end"] - scene_interferer["time_start"]
        interferer = select_random_interferer(
            interferers[scene_type], dataset, required_samples
        )
        scene_interferer["type"] = scene_type.value
        scene_interferer["name"] = interferer["ID"]
        scene_interferer["offset"] = get_random_interferer_offset(
            interferer, required_samples
        )


# listener handlinf


def get_random_hrir_set(heads, channels):
    """Get a random HRIR set."""
    head = random.choice(heads)
    hrir_names = [f"{head}-{channel}" for channel in channels]
    return hrir_names


def generate_rotation(
    scene,
    relative_start_time_range,
    duration_mean,
    duration_sd,
    angle_initial_mean,
    angle_initial_sd,
    angle_final_range,
):
    """Generate a suitable head rotation for the given scene.
    Based on behavioural studies by Hadley et al. TODO: find ref

    Args:
        scene (dict): the scene description
        head_turn_cfg (dict): head turn hyperparameters
    head_turn_cfg has the following fields:
        start_offset_mean - mean of the time offset for start of turn
        start_offset_sd - standard deviation of the time offset for start of turn
        angle_initial_mean
        angle_initial_sd
        angle_final_range (tuple)
        duration_mean - mean of the duration of the turn
        duration_sd - standard deviation of the duration of the turn

    Returns:
        list(dict): list of dicts with keys "sample" and "view_vector"
        specifying the head motion.
    """

    # Generate the random start and end time for the head rotation
    start_offset = random.uniform(*relative_start_time_range)
    duration = int(random.gauss(duration_mean, duration_sd))
    start_time = scene["target"]["time_start"] + start_offset
    start_time = max(start_time, 0)
    end_time = start_time + duration

    # Head initially offset to left or right at random
    random_sign = 1 if random.random() < 0.5 else -1

    # Generate the random start and end angle offsets for the head rotation
    offset = 0
    while True:
        offset = random.gauss(0, angle_initial_sd)
        # All rotations kept within 2 sigma of the mean
        if abs(offset) < 2 * angle_initial_sd:
            break

    angle_initial_offset = random_sign * math.radians(angle_initial_mean + offset)
    angle_final_offset = random_sign * (
        math.radians(random.uniform(*angle_final_range))
    )

    # Convert angle from relative to target speaker to room axes
    listener_pos = scene["room"]["listener"]["position"]
    target_pos = scene["room"]["target"]["position"]
    delta_x = target_pos[0] - listener_pos[0]
    delta_y = target_pos[1] - listener_pos[1]
    angle_target = math.atan2(delta_y, delta_x)

    # NOTE: JSON file will store angles in degrees
    angle_final = math.degrees(angle_target + angle_final_offset)
    angle_initial = math.degrees(angle_target + angle_initial_offset)

    # # Convert angles back to view vectors for the return object
    # start_view_vector = [math.cos(angle_initial), math.sin(angle_initial), 0.0]
    # end_view_vector = [math.cos(angle_final), math.sin(angle_final), 0.0]
    rotation = [
        {"sample": start_time, "angle": angle_initial},
        {"sample": end_time, "angle": angle_final},
    ]
    return rotation


class RoomBuilder:
    """Functions for handling rooms."""

    def __init__(self):
        """Initialize the room builder."""
        self.rooms = []
        self.room_dict = {}

    def get_room(self, name):
        """Get a room by name."""
        return self.room_dict[name]

    def rebuild_dict(self):
        """Build room dictionary."""
        self.room_dict = {room["name"]: room for room in self.rooms}

    def build_from_rpf(
        self, rpf_location, n_interferers=N_INTERFERERS, n_rooms=N_SCENES, start_room=1
    ):
        """Build a list of rooms by extracting info from RAVEN rpf files.

        Args:
            rpf_location (str): path to where rpf files are stored
            n_interferers (int, optional): number of interferer definitions to expect.
              Defaults to N_INTERFERERS.
            n_rooms (int, optional): number of scenes to expect. Defaults to N_SCENES.
            start_room (int, optional): index of the first room to expect
        """
        # Construct all rpf filenames for given number of interferers
        rpf_filedicts = [
            make_rpf_filename_dict(rpf_location, f"R{n:05d}", n_interferers)
            for n in range(start_room, start_room + n_rooms)
        ]

        # Process all rpf file to generate the list of rooms dicts
        self.rooms = [
            build_room(rpf_filedict["target"], rpf_filedict["interferers"])
            for rpf_filedict in tqdm(rpf_filedicts)
        ]
        self.rebuild_dict()

    def load(self, filename):
        """Load the list of room from a json file."""
        with open(filename, "r", encoding="utf-8") as f:
            self.rooms = json.load(f)
        self.rebuild_dict()

    def save_rooms(self, filename):
        """Save the list of rooms to a json file."""
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(self.rooms, f, indent=2)


class SceneBuilder:
    """Class with methods for building a list of scenes."""

    def __init__(
        self,
        rb,
        scene_datasets,
        target,
        interferer,
        snr_range,
        listener,
        shuffle_rooms=None,
    ):
        self.rb = rb
        self.scenes = []
        self.scene_datasets = scene_datasets
        self.target = target
        self.interferer = interferer
        self.snr_range = snr_range
        self.listener = listener
        self.shuffle_rooms = shuffle_rooms

    def save_scenes(self, filename):
        """Save the list of scenes to a json file."""
        scenes = [s for s in self.scenes]
        # Replace the room structure with the room ID
        for scene in scenes:
            scene["room"] = scene["room"]["name"]
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(self.scenes, f, indent=2)

    def instantiate_scenes(self, dataset):
        log.info("Initialise %s scenes", dataset)
        self.initialise_scenes(dataset, **self.scene_datasets)
        log.info("adding targets to scenes")
        self.add_target_to_scene(dataset, **self.target)
        log.info("adding interferers to scenes")
        self.add_interferer_to_scene(**self.interferer)
        log.info("assigning an SNR to each scene")
        self.add_SNR_to_scene(self.snr_range)
        log.info("adding listener behaviours to scenes")
        self.add_listener_details_to_scene(**self.listener)

    def initialise_scenes(self, dataset, n_scenes, room_selection, scene_start_index):
        """
        Initialise the scenes for a given dataset.

        Args:
            dataset: train, dev, or eval set
            n_scenes: number of scenes to generate
            room_selection: SEQUENTIAL or RANDOM
            scene_start_index: index to start for scene IDs
        """
        rooms = self.rb.rooms.copy()
        if self.shuffle_rooms:
            random.shuffle(rooms)

        # Construct the scenes adding the room and dataset label
        self.scenes = []
        scenes = [{"dataset": dataset} for _ in range(n_scenes)]
        # Sequential mode: cycle through all available rooms
        if room_selection == "SEQUENTIAL":
            # Cycle through rooms if not enough for required number of scenes
            for scene, room in zip(scenes, itertools.cycle(rooms)):
                scene["room"] = room
        # Random mode: randomly select rooms without replacement
        elif room_selection == "RANDOM":
            for scene in scenes:
                scene["room"] = random.choice(rooms)
        else:
            assert False, "Unknown room selection mode"
        self.scenes.extend(scenes)

        # Set the scene ID
        for index, scene in enumerate(self.scenes, scene_start_index):
            scene["scene"] = f"S{index:05d}"

    def add_target_to_scene(
        self,
        dataset,
        target_speakers,
        target_selection,
        pre_samples_range,
        post_samples_range,
    ):
        """Add target info to the scenes.
        Target speaker file set via config.

        Raises:
            Exception: _description_
        """
        with open(target_speakers, "r", encoding="utf-8") as f:
            targets = json.load(f)

        targets_dataset = [t for t in targets if t["dataset"] == dataset]
        scenes_dataset = [s for s in self.scenes if s["dataset"] == dataset]

        random.shuffle(targets_dataset)

        if target_selection == "SEQUENTIAL":
            # Sequential mode: Cycle through targets sequentially
            for scene, target in zip(scenes_dataset, itertools.cycle(targets_dataset)):
                add_this_target_to_scene(
                    target, scene, pre_samples_range, post_samples_range
                )
        elif target_selection == "RANDOM":
            # Random mode: randomly select target with replacement
            for scene in scenes_dataset:
                add_this_target_to_scene(
                    random.choice(targets_dataset),
                    scene,
                    pre_samples_range,
                    post_samples_range,
                )
        else:
            assert False, "Unknown target selection mode"

    def add_SNR_to_scene(self, snr_range):
        """Add the SNR info to the scenes."""
        for scene in tqdm(self.scenes):
            scene["SNR"] = generate_snr(snr_range)

    def add_interferer_to_scene(
        self,
        speech_interferers,
        noise_interferers,
        music_interferers,
        number,
        start_time_range,
        end_early_time_range,
    ):
        """Add interferer to the scene description file."""
        # Load and prepare speech interferer metadata
        with open(speech_interferers, "r", encoding="utf-8") as f:
            interferers_speech = json.load(f)
        for interferer in interferers_speech:
            interferer["ID"] = (
                interferer["speaker"] + ".wav"
            )  # selection require a unique "ID" field
        # Selection process requires list of lists
        interferers_speech = [interferers_speech]

        # Load and prepare noise (i.e. noise) interferer metadata
        with open(noise_interferers, "r", encoding="utf-8") as f:
            interferers_noise = json.load(f)
        for interferer in interferers_noise:
            interferer["ID"] += ".wav"
        interferer_by_type = {}
        for interferer in interferers_noise:
            interferer_by_type.setdefault(interferer["class"], []).append(interferer)
        interferers_noise = list(interferer_by_type.values())

        # Load and prepare music interferer metadata
        with open(music_interferers, "r", encoding="utf-8") as f:
            interferers_music = json.load(f)
        for interferer in interferers_music:
            interferer["ID"] = interferer[
                "file"
            ]  # selection require a unique "ID" field
        interferers_music = [interferers_music]

        interferers = {
            InterfererType.SPEECH: interferers_speech,
            InterfererType.NOISE: interferers_noise,
            InterfererType.MUSIC: interferers_music,
        }

        for scene in tqdm(self.scenes):
            add_interferer_to_scene_inner(
                scene, interferers, number, start_time_range, end_early_time_range
            )

    def add_listener_details_to_scene(
        self,
        heads,
        channels,
        relative_start_time_range,
        duration_mean,
        duration_sd,
        angle_initial_mean,
        angle_initial_sd,
        angle_final_range,
    ):
        """Add the listener info to the scenes."""
        for scene in tqdm(self.scenes):
            listener = {}
            listener["rotation"] = generate_rotation(
                scene,
                relative_start_time_range,
                duration_mean,
                duration_sd,
                angle_initial_mean,
                angle_initial_sd,
                angle_final_range,
            )
            listener["hrir_filename"] = get_random_hrir_set(heads, channels)
            scene["listener"] = listener



import torch

from clarity.predictor.torch_stoi import NegSTOILoss


class SISNRLoss(torch.nn.Module):
    # def __init__(self):  # removed as does nothing
    #    super(SISNRLoss, self).__init__()

    def cal_sisnr(self, x, s, eps=1e-8):
        """
        Arguments:
        x: separated signal, N x S tensor
        s: reference signal, N x S tensor
        Return:
        sisnr: N tensor
        """

        def l2norm(mat, keepdim=False):
            return torch.norm(mat, dim=-1, keepdim=keepdim)

        if x.shape != s.shape:
            raise RuntimeError(
                f"Dimension mismatch when calculate si-snr, {x.shape} vs {s.shape}"
            )
        x_zm = x - torch.mean(x, dim=-1, keepdim=True)
        s_zm = s - torch.mean(s, dim=-1, keepdim=True)
        t = (
            torch.sum(x_zm * s_zm, dim=-1, keepdim=True)
            * s_zm
            / (l2norm(s_zm, keepdim=True) ** 2 + eps)
        )
        return 20 * torch.log10(eps + l2norm(t) / (l2norm(x_zm - t) + eps))

    def forward(self, x, y):
        return -self.cal_sisnr(x, y).mean()


class SNRLoss(torch.nn.Module):
    def __init__(self, tao=1e-3):
        super().__init__()
        self.tao = tao

    def l2norm(self, mat, keepdim=False):
        return torch.norm(mat, dim=-1, keepdim=keepdim)

    def forward(self, x, s, eps=1e-8):
        if x.shape != s.shape:
            raise RuntimeError(
                f"Dimension mismatch when calculate si-snr, {x.shape} vs {s.shape}"
            )

        loss = 10 * torch.log10(
            self.l2norm(s - x) ** 2 + self.tao * self.l2norm(s) ** 2 + eps
        ) - 10 * torch.log10(self.l2norm(s) ** 2 + eps)
        return loss.mean()


class STOILoss(torch.nn.Module):
    def __init__(self, sr):
        super().__init__()
        self.NegSTOI = NegSTOILoss(sample_rate=sr)

    def forward(self, x, s):
        return self.NegSTOI(x, s).mean()


class STOILevelLoss(torch.nn.Module):
    def __init__(self, sr, alpha, block_size=0.4, overlap=0.7, gamma_a=-70):
        super().__init__()
        self.NegSTOI = NegSTOILoss(sample_rate=sr)
        self.alpha = alpha

        "rms measurement"
        self.frame_size = int(block_size * sr)
        self.frame_shift = int(block_size * sr * (1 - overlap))
        self.unfold = torch.nn.Unfold(
            (1, self.frame_size), stride=(1, self.frame_shift)
        )
        self.gamma_a = gamma_a

        "mse"
        self.cal_mse = torch.nn.MSELoss()

    def measure_loudness(self, signal, eps=1e-8):
        x_unfold = self.unfold(signal.unsqueeze(1).unsqueeze(2))

        z = (
            torch.sum(x_unfold**2, dim=1) / self.frame_size
        )  # mean square for each frame
        el = -0.691 + 10 * torch.log10(z + eps)

        idx_a = torch.where(el > self.gamma_a, 1, 0)
        z_ave_gated_a = torch.sum(z * idx_a, dim=1, keepdim=True) / (
            torch.sum(idx_a, dim=1, keepdim=True) + eps
        )
        gamma_r = -0.691 + 10 * torch.log10(z_ave_gated_a + eps) - 10

        idx_r = torch.where(el > gamma_r, 1, 0)
        idx_a_r = idx_a * idx_r
        z_ave_gated_a_r = torch.sum(z * idx_a_r, dim=1, keepdim=True) / (
            torch.sum(idx_a_r, dim=1, keepdim=True) + eps
        )
        lufs = -0.691 + 10 * torch.log10(z_ave_gated_a_r + eps)  # loudness
        return lufs

    def forward(self, x, s):
        loudness_x = self.measure_loudness(x)
        loudness_s = self.measure_loudness(s)
        LevelLoss = self.alpha * self.cal_mse(loudness_x, loudness_s)
        return LevelLoss + self.NegSTOI(x, s).mean()



"""Matlab's haaqi version 1 to python version."""
import numpy as np

from clarity.evaluator.haspi import eb


def haaqi_v1(
    x: np.ndarray,
    fx: int,
    y: np.ndarray,
    fy: int,
    hl: np.ndarray,
    eq: int,
    level1: int = 65,
):
    """
    Compute the HAAQI music quality index using the
    auditory model followed by computing the envelope cepstral
    correlation and BM vibration average short-time coherence signals.
    The reference signal presentation level for NH listeners is assumed
    to be 65 dB SPL. The same model is used for both normal and
    impaired hearing.

    Arguments:
        x (ndarray):  Clear input reference speech signal with no noise or distortion.
            If a hearing loss is specified, NAL-R equalization is optional
        fx: Sampling rate in Hz for signal x
        y:  Output signal with noise, distortion, HA gain, and/or processing.
        fy: Sampling rate in Hz for signal y.
        hl: (1,6) vector of hearing loss at the 6 audiometric frequencies
        [250, 500, 1000, 2000, 4000, 6000] Hz.
        eq: Flag to provide equalization for the hearing loss to signal x:
            1 = no EQ has been provided, the function will add NAL-R
            2 = NAL-R EQ has already been added to the reference signal
        level1: Optional input specifying level in dB SPL that corresponds to a
           signal RMS = 1. Default is 65 dB SPL if argument not provided.
           Default: 65

    Returns:
        combined : Quality is the polynomial sum of the nonlin and linear terms
        nonlin : Nonlinear quality component = .245(BMsync5) + .755(CepHigh)^3
        linear : Linear quality component = std of spectrum and norm spectrum
        raw : Vector of raw values = [cephigh, bmsync5, dloud, dnorm]

    James M. Kates, 5 August 2013 (HASQI_v2).
    Version for HAAQI_v1, 19 Feb 2015.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Auditory model for quality
    # Reference is no processing or NAL-R, impaired hearing
    xenv, xbm, yenv, ybm, xsl, ysl, fsamp = eb.EarModel(x, fx, y, fy, hl, eq, level1)

    # ---------------------------------------
    # Envelope and long-term average spectral features
    # Smooth the envelope outputs: 250 Hz sub-sampling rate
    segsize = 8  # Averaging segment size in msec
    xdb = eb.env_smooth(xenv, segsize, fsamp)
    ydb = eb.env_smooth(yenv, segsize, fsamp)

    # Mel cepstrum correlation after passing through modulation filterbank
    thr = 2.5  # Silence threshold: sum across bands, dB above aud threshold
    addnoise = 0.0  # dditive noise in dB SL to condition cross-covariances
    _, _, cephigh, _ = eb.melcor9(
        xdb, ydb, thr, addnoise, segsize
    )  # 8 modulation freq bands

    # Linear changes in the long-term spectra
    # dloud  vector: [sum abs diff, std dev diff, max diff] spectra
    # dnorm  vector: [sum abs diff, std dev diff, max diff] norm spectra
    # dslope vector: [sum abs diff, std dev diff, max diff] slope
    dloud_vector, dnorm_vector, _ = eb.spect_diff(xsl, ysl)

    # Temporal fine structure (TFS) correlation measurements
    # Compute the time-frequency segment covariances
    segcov = 16  # Segment size for the covariance calculation
    sigcov, sigmsx, _ = eb.bm_covary(xbm, ybm, segcov, fsamp)

    # Average signal segment cross-covariance
    # avecov=weighted ave of cross-covariances, using only data above threshold
    # syncov=ave cross-covariance with added IHC loss of synchronization at HF
    thr = 2.5  # Threshold in dB SL for including time-freq tile
    _, syncov = eb.ave_covary2(sigcov, sigmsx, thr)
    bmsync5 = syncov[4]  # Ave segment coherence with IHC loss of sync

    # Extract and normalize the spectral features
    # Dloud:std
    d = dloud_vector[1]  # Loudness difference std
    d = d / 2.5  # Scale the value
    d = 1.0 - d  # 1=perfect, 0=bad
    d = min(d, 1)
    d = max(d, 0)
    dloud = d

    # Dnorm:std
    d = dnorm_vector[1]  # Slope difference std
    d = d / 25  # Scale the value
    d = 1.0 - d  # 1=perfect, 0=bad
    d = min(d, 1)
    d = max(d, 0)
    dnorm = d

    # Construct the models
    # Nonlinear model
    nonlin_model = 0.754 * (cephigh**3) + 0.246 * bmsync5  # Combined envelope and TFS

    # Linear model
    lin_model = 0.329 * dloud + 0.671 * dnorm  # Linear fit

    # Combined model
    combined = (
        0.336 * nonlin_model
        + 0.001 * lin_model
        + 0.501 * (nonlin_model**2)
        + 0.161 * (lin_model**2)
    )  # Polynomial sum

    # Raw data
    raw = [cephigh, bmsync5, dloud, dnorm]

    return combined, nonlin_model, lin_model, raw



import copy
import json
import logging
import os

import hydra
import speechbrain as sb
import torch
from hyperpyyaml import load_hyperpyyaml
from omegaconf import DictConfig
from speechbrain.utils.distributed import run_on_main
from tqdm import tqdm
from transformer_cpc1_ensemble_decoder import S2STransformerBeamSearch

logger = logging.getLogger(__name__)


class ASR(sb.core.Brain):
    def compute_uncertainty(self, wavs, wav_lens, tokens_bos):
        """Forward computations from the waveform batches to the output probabilities."""
        # batch = batch.to(self.device)
        wavs, wav_lens, tokens_bos = (
            wavs.to(self.device),
            wav_lens.to(self.device),
            tokens_bos.to(self.device),
        )
        with torch.no_grad():
            feats = self.hparams.compute_features(wavs)
            current_epoch = self.hparams.epoch_counter.current
            feats = self.hparams.normalize(feats, wav_lens, epoch=current_epoch)

            cnn_out = []
            for j in range(self.hparams.n_ensembles):
                cnn_out.append(self.asr_ensemble[j][0](feats))
            _, _, prob_outputs = self.test_search(
                cnn_out, tokens_bos, wav_lens, pad_idx=self.hparams.pad_index
            )

        return prob_outputs

    def init_ensembles(self, n_ensemble):
        ensembles = []
        for j in range(n_ensemble):
            ensembles.append(copy.deepcopy(self.hparams.model))
        return ensembles

    def init_evaluation(self, max_key=None, min_key=None):
        """perform checkpoint averge if needed"""
        super().on_evaluate_start()

        self.asr_ensemble = self.init_ensembles(self.hparams.n_ensembles)
        ckpts = self.checkpointer.find_checkpoints(max_key=max_key, min_key=min_key)
        test_search_modules = []
        for j in range(self.hparams.n_ensembles):
            ckpt = sb.utils.checkpoints.average_checkpoints(
                [ckpts[j]], recoverable_name="model", device=self.device
            )
            self.asr_ensemble[j].load_state_dict(ckpt, strict=True)
            self.asr_ensemble[j].eval()
            test_search_modules.extend(self.asr_ensemble[j][1:])

        self.test_search = S2STransformerBeamSearch(
            modules=test_search_modules,
            n_ensembles=self.hparams.n_ensembles,
            bos_index=self.hparams.bos_index,
            eos_index=self.hparams.eos_index,
            blank_index=self.hparams.blank_index,
            min_decode_ratio=self.hparams.min_decode_ratio,
            max_decode_ratio=self.hparams.max_decode_ratio,
            beam_size=self.hparams.test_beam_size,
            ctc_weight=self.hparams.ctc_weight_decode,
            lm_weight=self.hparams.lm_weight,
            lm_modules=self.hparams.lm_model,
            temperature=self.hparams.temperature,
            temperature_lm=1,
            topk=self.hparams.topk,
            using_eos_threshold=False,
            length_normalization=True,
        )


def init_asr(asr_config):
    hparams_file, run_opts, overrides = sb.parse_arguments([asr_config])
    with open(hparams_file) as fin:
        hparams = load_hyperpyyaml(fin, overrides)

    tokenizer = hparams["tokenizer"]
    bos_index = hparams["bos_index"]

    # We download the pretrained LM from HuggingFace (or elsewhere depending on
    # the path given in the YAML file). The tokenizer is loaded at the same time.
    run_on_main(hparams["pretrainer"].collect_files)
    hparams["pretrainer"].load_collected(device=run_opts["device"])

    asr_brain = ASR(
        modules=hparams["modules"],
        opt_class=hparams["Adam"],
        hparams=hparams,
        run_opts=run_opts,
        checkpointer=hparams["checkpointer"],
    )
    asr_brain.init_evaluation()

    return asr_brain, tokenizer, bos_index


def compute_uncertainty(left_proc_path, asr_model, bos_index, tokenizer):
    wav_len = torch.tensor([1], dtype=torch.float32)
    tokens_bos = torch.LongTensor([bos_index]).view(1, -1)

    right_proc_path = left_proc_path.replace("left", "right")
    left_proc_wav = sb.dataio.dataio.read_audio(left_proc_path).view(1, -1)
    right_proc_wav = sb.dataio.dataio.read_audio(right_proc_path).view(1, -1)

    left_uncertainty = asr_model.compute_uncertainty(left_proc_wav, wav_len, tokens_bos)
    right_uncertainty = asr_model.compute_uncertainty(
        right_proc_wav, wav_len, tokens_bos
    )
    conf = max(
        left_uncertainty[0]["confidence"].detach().cpu().numpy(),
        right_uncertainty[0]["confidence"].detach().cpu().numpy(),
    )
    neg_ent = -min(
        left_uncertainty[0]["entropy"].detach().cpu().numpy(),
        right_uncertainty[0]["entropy"].detach().cpu().numpy(),
    )
    return conf, neg_ent


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    if cfg.cpc1_track == "open":
        track = "_indep"
    elif cfg.cpc1_track == "closed":
        track = ""
    else:
        logger.error("cpc1_track has to be closed or open")

    asr_model, tokenizer, bos_index = init_asr(cfg.asr_config)

    left_dev_csv = sb.dataio.dataio.load_data_csv(
        os.path.join(cfg.path.exp_folder, "cpc1_asr_data" + track, "left_dev_msbg.csv")
    )  # using left ear csvfile for data loading
    left_test_csv = sb.dataio.dataio.load_data_csv(
        os.path.join(cfg.path.exp_folder, "cpc1_asr_data" + track, "left_test_msbg.csv")
    )  # using left ear csvfile for data loading

    dev_conf = {}
    dev_negent = {}
    # dev set uncertainty
    for wav_id, wav_obj in tqdm(left_dev_csv.items()):
        left_proc_path = wav_obj["wav"]
        uncertainty = compute_uncertainty(
            left_proc_path, asr_model, bos_index, tokenizer
        )
        dev_conf[wav_id] = uncertainty[0].tolist()
        dev_negent[wav_id] = uncertainty[1].tolist()

        with open(os.path.join(cfg.path.exp_folder, "dev_conf.json"), "w") as f:
            json.dump(dev_conf, f)
        with open(os.path.join(cfg.path.exp_folder, "dev_negent.json"), "w") as f:
            json.dump(dev_negent, f)

    # test set similarity
    test_conf = {}
    test_negent = {}
    for wav_id, wav_obj in tqdm(left_test_csv.items()):
        left_proc_path = wav_obj["wav"]
        uncertainty = compute_uncertainty(
            left_proc_path, asr_model, bos_index, tokenizer
        )
        test_conf[wav_id] = uncertainty[0].tolist()
        test_negent[wav_id] = uncertainty[1].tolist()

        with open(os.path.join(cfg.path.exp_folder, "test_conf.json"), "w") as f:
            json.dump(test_conf, f)
        with open(os.path.join(cfg.path.exp_folder, "test_negent.json"), "w") as f:
            json.dump(test_negent, f)


if __name__ == "__main__":
    run()
